// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
import KeychainAccess

@dynamicMemberLookup
class BubbleKeychain {
    class KeyHolder<T> {
        let name: String
        var value: T?
        
        init(name: String) {
            self.name = name
        }
    }
    struct KeyStore {
        var currentBubbleManager: KeyHolder<BubbleManager> { .init(name: #function) }
        var currentSageManager: KeyHolder<SageManager> { .init(name: #function) }
        
        #if os(macOS)
        var flexrouterPassword: KeyHolder<String> { .init(name: #function) }
        #endif
    }
    
    static let shared = BubbleKeychain(keychain: .init(service: "com.bubble.bubble-vpn.credentials"))
    
    private let keychain: KeychainAccess.Keychain
    private var keyStore = KeyStore()
    
    init(keychain: KeychainAccess.Keychain) {
        self.keychain = keychain
    }
    
    func removeNetworkKeys() {
        self.currentBubbleManager = nil
        self.currentSageManager = nil
    }
    
    subscript<T: Codable>(dynamicMember keyPath: KeyPath<KeyStore, KeyHolder<T>>) -> T? {
        get {
            if T.self == String.self {
                return keychain[keyStore[keyPath: keyPath].name] as? T
            }
            return try? keychain[data: keyStore[keyPath: keyPath].name]?.decode(T.self)
        }
        set {
            if T.self == String.self {
                keychain[keyStore[keyPath: keyPath].name] = newValue as? String
            } else {
                if let newValue = newValue {
                    keychain[data: keyStore[keyPath: keyPath].name] = try? JSONEncoder().encode(newValue)
                } else {
                    try? keychain.remove(keyStore[keyPath: keyPath].name)
                }
            }
        }
    }
}

extension Data {
    func decode<T: Decodable>(_ type: T.Type) throws -> T {
        try JSONDecoder().decode(T.self, from: self)
    }
}
