// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
import SwiftyUserDefaults

class ConnectionManager: NSObject {
    enum Status: Equatable {
        case initial
        
        case noPaymentMethod
        case noBubble
        case startingBubble(progress: Double)
        case inactive
        case activating
        case active
        case deactivating
        case restarting
        
        init(tunnelStatus: TunnelStatus) {
            switch tunnelStatus {
            case .inactive:
                self = .inactive
                
            case .activating:
                self = .activating
                
            case .active:
                self = .active
                
            case .deactivating:
                self = .deactivating
                
            case .restarting:
                self = .restarting
                
            case .reasserting, .waiting:
                self = .inactive
            }
        }
        
        var localizedStatusDescription: String {
            switch self {
            case .initial, .inactive, .startingBubble, .noBubble, .noPaymentMethod:
                return tr("tunnelStatusInactive")
            case .activating:
                return tr("tunnelStatusActivating")
            case .active:
                return tr("tunnelStatusActive")
            case .deactivating:
                return tr("tunnelStatusDeactivating")
            case .restarting:
                return tr("tunnelStatusRestarting")
            }
        }
        
        var localizedToggleStatusActionText: String {
            switch self {
            case .initial, .inactive, .startingBubble, .noBubble, .noPaymentMethod:
                return tr("macToggleStatusButtonActivate")
            case .activating:
                return tr("macToggleStatusButtonActivating")
            case .active:
                return tr("macToggleStatusButtonDeactivate")
            case .deactivating:
                return tr("macToggleStatusButtonDeactivating")
            case .restarting:
                return tr("macToggleStatusButtonRestarting")
            }
        }
    }
    
    static let tunnelName = "BubbleVPN"
    
    var autoConnectOnStart = true
    
    class StatusHandler {
        fileprivate let handler: (Status) -> Void
        
        init(_ handler: @escaping (Status) -> Void) {
            self.handler = handler
        }
    }
    
    private var statusChangedHandlers = NSHashTable<StatusHandler>.weakObjects()
    
    func addStatusChangedHandler(statusHandler: StatusHandler) -> StatusHandler {
        statusChangedHandlers.add(statusHandler)
        return statusHandler
    }
    
    private(set) var status: Status = .initial {
        didSet {
            guard status != oldValue else { return }
            if status == .active {
                validateSSL()
            } else {
                sslValidationCancellable = nil
            }
            #if os(macOS)
            if status == .active {
                registerFlexrouterIfNeeded()
            } else {
                unregisterFlexrouter()
            }
            #endif
            runOnMainThread {
                self.statusChangedHandlers.allObjects.forEach { $0.handler(self.status) }   
            }
        }
    }
    @objc private(set) dynamic var error: Error?
    
    private var tunnelsManager: TunnelsManager! {
        didSet {
            guard tunnelsManager != nil else {
                tunnelContainer = nil
                status = .initial
                return
            }
            status = .inactive
            tunnelsManager.activationDelegate = self
            tunnelsManager.tunnelsListDelegate = self
            tunnelContainer = tunnelsManager.tunnel(named: Self.tunnelName)
            if autoConnectOnStart {
                connect()
            }
        }
    }
    private var observationToken: AnyObject?
    private weak var tunnelContainer: TunnelContainer? {
        didSet {
            observationToken = tunnelContainer?.observe(\.status) { [weak self] tunnel, _ in
                self?.status = .init(tunnelStatus: tunnel.status)
            }
            if let tunnelStatus = tunnelContainer?.status {
                status = .init(tunnelStatus: tunnelStatus)
            }
        }
    }
    private let sageManager: SageManager
    private var bubbleManager: BubbleManager?
    
    var myBubbleURL: URL? { bubbleManager?.myBubbleURL ?? sageManager.myBubbleURL }
    var accountURL: URL? { bubbleManager?.accountURL ?? sageManager.accountURL }
    var launchBubbleURL: URL? { sageManager.launchBubbleURL }
    var paymentPageURL: URL? { sageManager.paymentPageURL }
    
    init(sageManager: SageManager) {
        self.sageManager = sageManager
        bubbleManager = BubbleKeychain.shared.currentBubbleManager
    }
    
    func initialize() {
        guard status == .initial else { return }
        createTunnelsManager()
    }
    
    func logOut() {
        if let tunnelContainer = tunnelContainer {
            tunnelsManager.remove(tunnel: tunnelContainer) { _ in }
        }
        sageManager.destroy()
        bubbleManager?.destroy()
        BubbleKeychain.shared.removeNetworkKeys()
        #if os(macOS)
        clearCert()
        unregisterFlexrouter()
        #endif
    }
    
    func toggleStatus() {
        switch status {
        case .inactive:
            connect()
        default:
            disconnect()
        }
    }
    
    func connect() {
        error = nil
        if let tunnelContainer = tunnelContainer {
            tunnelsManager.startActivation(of: tunnelContainer)
            return
        }
        status = .activating
        updateBubbleManager()
    }
    
    func disconnect() {
        error = nil
        if let tunnelContainer = tunnelContainer {
            tunnelsManager.startDeactivation(of: tunnelContainer)
        } else {
            installVPNCancellable = nil
        }
    }
    
    private func createTunnelsManager() {
        TunnelsManager.create { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .failure(let error):
                self.error = error
            case .success(let tunnelsManager):
                self.tunnelsManager = tunnelsManager
            }
        }
    }
    
    var installVPNCancellable: Cancellable? {
        didSet {
            oldValue?.cancel()
        }
    }
    
    private func installVPNConfig() {
        installVPNCancellable = bubbleManager?.getVPNConf { [weak self] result in
            let handleError = { [weak self] (error: Error) in
                guard let self = self else { return }
                self.error = error
                self.status = .inactive
            }
            do {
                switch result {
                case .success(let url):
                    let fileContents: String
                    fileContents = try String(contentsOf: url)
                    let tunnelConfiguration = try TunnelConfiguration(fromWgQuickConfig: fileContents,
                                                                      called: Self.tunnelName)
                    self?.tunnelsManager.add(tunnelConfiguration: tunnelConfiguration) { [weak self] result in
                        guard let self = self else { return }
                        switch result {
                        case .success(let container):
                            self.tunnelContainer = container
                            self.tunnelsManager.startActivation(of: container)
                            
                        case .failure(let error):
                            handleError(error)
                        }
                    }
                    
                case .failure(let error):
                    throw error
                }
            } catch {
                handleError(error)
            }
        }
    }
    
    #if os(iOS)
    var caCertURL: URL? {
        bubbleManager?.caCertURL
    }
    var invalidSSLHandler: (() -> Void)?
    #elseif os(macOS)
    private var needsInstallCACertificate = false
    
    private func setNeedsInstallCACertificate() {
        guard needsInstallCACertificate == false else { return }
        needsInstallCACertificate = true
        error = nil
        bubbleManager?.getCACertificate { [weak self] result in
            guard let self = self else { return }
            self.needsInstallCACertificate = false
            do {
                switch result {
                case .success(let url):
                    try self.addAndTrustRootCert(url: url)
                    
                case .failure(let error):
                    throw error
                }
            } catch {
                self.error = error
            }
        }
    }
    
    private func addAndTrustRootCert(url: URL) throws {
        let pemRegex = try NSRegularExpression(pattern: "-----BEGIN CERTIFICATE-----([^-]+)-----END CERTIFICATE-----", options: .caseInsensitive)
        let clearString = try String(contentsOf: url).replacingOccurrences(of: "\n", with: "")
        guard
            let match = pemRegex.firstMatch(in: clearString,
                                            range: .init(location: 0, length: clearString.count)),
            match.numberOfRanges == 2,
            case let derString = (clearString as NSString).substring(with: match.range(at: 1)),
            let rootCertData = Data(base64Encoded: derString),
            let rootCert = rootCertData.createSecCertificate()
            else {
                throw NSError(description: "Invalid certificate")
        }
        
        let oldCertData = Defaults.installedRootCert
        if oldCertData != rootCertData {
            switch SecItemAdd(
                [kSecClass: kSecClassCertificate,
                 kSecValueRef: rootCert,
                    ] as CFDictionary,
                nil) {
            case errSecSuccess, errSecDuplicateItem:
                break
                
            case let status:
                throw NSError(description: "Adding root certificate error", code: .init(status))
            }
            Defaults.installedRootCert = rootCertData
        }
        
        var trustUpdateNeeded = true
        var trustSettings: CFArray!
        if SecTrustSettingsCopyTrustSettings(rootCert, .user, &trustSettings) == 0,
            let trustSettings = trustSettings as? [[String: Any]] {
            trustUpdateNeeded = false
            for setting in trustSettings {
                guard
                    let trustSettingsResultRaw = setting[kSecTrustSettingsResult] as? UInt32,
                    let trustSettingsResult = SecTrustSettingsResult(rawValue: trustSettingsResultRaw)
                    else { break }
                switch trustSettingsResult {
                case .trustAsRoot, .trustRoot:
                    continue
                    
                case .deny, .invalid, .unspecified:
                    trustUpdateNeeded = true
                @unknown default:
                    trustUpdateNeeded = true
                }
                break
            }
        }
        guard trustUpdateNeeded else { return }
        try tryOSStatus(
            SecTrustSettingsSetTrustSettings(rootCert, .user, nil),
            description: "Trusting root certificate"
        )
    }
    
    private func clearCert() {
        do {
            try clearCertThrows()
        } catch {
            wg_log(error: error, message: #function)
        }
    }
    
    private func clearCertThrows() throws {
        guard let oldCert = Defaults.installedRootCert?.createSecCertificate() else { return }
        Defaults.installedRootCert = nil
        var res: CFTypeRef?
        try tryOSStatus(
            SecItemCopyMatching(
                [kSecClass: kSecClassCertificate,
                 kSecValueRef: oldCert,
                    ] as CFDictionary,
                &res),
            description: #function + " SecItemCopyMatching"
        )
        try tryOSStatus(
            SecItemDelete(
                [kSecClass: kSecClassCertificate,
                 kSecMatchItemList: [res],
                    ] as CFDictionary),
            description: #function + " SecItemDelete"
        )
    }
    
    var requestFlexrouterPassword: ((FlexrouterManager.Error?) -> Void)?
    private var flexrouterCancellable: Cancellable? { didSet { oldValue?.cancel() } }
    private var flexrouterManager: FlexrouterManager?
    
    func updateFlexrouterPassword(
        _ newPassword: String,
        completionHandler: @escaping (Result<Void, FlexrouterManager.Error>) -> Void
    ) {
        registerFlexrouter(
            newPassword,
            updated: true,
            completionHandler: completionHandler
        )
    }
    
    private func registerFlexrouter(
        _ password: String,
        updated: Bool,
        completionHandler: ((Result<Void, FlexrouterManager.Error>) -> Void)? = nil
    ) {
        guard
            flexrouterCancellable == nil,
            status == .active,
            let bubbleManager = bubbleManager,
            let sessionToken = bubbleManager.sessionToken
        else { return }
        flexrouterCancellable = bubbleManager.getCurrentDevice { [weak self] result in
            guard let self = self else { return }
            self.flexrouterCancellable = nil
            guard let vpnIp = (try? result.get())?.status?.vpnIps.first else {
                wg_log(.error, message: "current device info invalid: \(result)")
                DispatchQueue.main.asyncAfter(deadline: .now() + 1) { [weak self] in
                    self?.registerFlexrouter(
                        password,
                        updated: updated,
                        completionHandler: completionHandler
                    )
                }
                return
            }
            self.flexrouterCancellable = self.flexrouterManager?.register(
                .init(password: password,
                      session: sessionToken,
                      bubble: bubbleManager.host,
                      ip: vpnIp))
            { [weak self] result in
                guard let self = self else { return }
                self.flexrouterCancellable = nil
                guard case .failure(let error) = result else {
                    if updated {
                        BubbleKeychain.shared.flexrouterPassword = password
                    }
                    completionHandler?(.success(()))
                    wg_log(.info, message: "flexrouter registered")
                    return
                }
                if let completionHandler = completionHandler {
                    completionHandler(.failure(error))
                } else {
                    self.requestFlexrouterPassword?(error)
                }
            }
        }
    }
    
    private func unregisterFlexrouter() {
        flexrouterCancellable = nil
        guard
            let flexrouterManager = flexrouterManager,
            let flexrouterPassword = BubbleKeychain.shared.flexrouterPassword
        else { return }
        flexrouterCancellable = flexrouterManager
            .unregister(password: flexrouterPassword) { [weak self] result in
                switch result {
                case .success:
                    wg_log(.info, message: "flexrouter unregistered")
                    self?.flexrouterManager = nil
                    
                case .failure(let error):
                    wg_log(error: error, message: "flexrouter unregister failed")
                }
            }
    }
    
    private func registerFlexrouterIfNeeded() {
        guard flexrouterCancellable == nil else { return }
        if flexrouterManager == nil {
            flexrouterManager = .init()
        }
        flexrouterCancellable = flexrouterManager?.ping { [weak self] result in
            guard let self = self else { return }
            self.flexrouterCancellable = nil
            guard case .success = result else {
                wg_log(.info, message: "flexrouter not detected")
                self.flexrouterManager = nil
                return
            }
            if let password = BubbleKeychain.shared.flexrouterPassword {
                self.registerFlexrouter(password, updated: false)
            } else {
                self.requestFlexrouterPassword?(nil)
            }
        }
    }
    #endif
    
    private var sslValidationCancellable: Cancellable? {
        didSet {
            oldValue?.cancel()
        }
    }
    
    private func validateSSL() {
        #if os(macOS)
        if Defaults.installedRootCert == nil {
            setNeedsInstallCACertificate()
        }
        #elseif os(iOS)
        sslValidationCancellable = bubbleManager?.validateSSL { [weak self] in
            guard let self = self else { return }
            self.invalidSSLHandler?()
        }
        #endif
    }
    
    private func getBubbleManager(completionHandler: @escaping (BubbleManager) -> Void) {
        if let bubbleManager = bubbleManager {
            completionHandler(bubbleManager)
            return
        }
        sageManager.getBubbleManager(.init(bubbleStartingStatusHandler: { [weak self] status in
            self?.status = .startingBubble(progress: status.percent / 100)
        }, completionHandler: { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .failure(let error):
                if error == .noBubble {
                    self.checkPaymentMethod()
                } else {
                    self.status = .inactive
                    self.error = error
                }
                
            case .success(let bubbleManager):
                self.bubbleManager = bubbleManager
                BubbleKeychain.shared.currentBubbleManager = bubbleManager
                completionHandler(bubbleManager)
            }
        }))
    }
    
    private func updateBubbleManager() {
        getBubbleManager { [weak self] _ in
            self?.status = .activating
            self?.installVPNConfig()
        }
    }
    
    func checkPaymentMethod() {
        sageManager.hasValidPaymentMethod { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .success(let hasValidPaymentMethod):
                if hasValidPaymentMethod {
                    self.createAccountPlanAndLaunch()
                } else {
                    self.status = .noPaymentMethod
                }
                
            case .failure(let error):
                self.status = .inactive
                self.error = error
            }
        }
    }
    
    func createAccountPlanAndLaunch() {
        status = .activating
        sageManager.createAccountPlan { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .success(let accountPlan):
                self.sageManager.launchBubble(accountPlan) { [weak self] result in
                    guard let self = self else { return }
                    switch result {
                    case .success:
                        self.updateBubbleManager()
                        
                    case .failure(let error):
                        self.status = .inactive
                        self.error = error
                    }
                }
                
            case .failure(let error):
                self.status = .inactive
                self.error = error
            }
        }
    }
    
    struct Card {
        var number: String
        var expiryMonth: Int
        var expiryYear: Int
        var cvc: Int
        var addressZip: Int
        
        var stripeRequestParameters: [String: Any] {
            ["card[number]": number,
             "card[exp_month]": expiryMonth,
             "card[exp_year]": expiryYear,
             "card[cvc]": cvc,
             "card[address_zip]": addressZip]
        }
        static var empty = Self(number: "", expiryMonth: 0, expiryYear: 0, cvc: 0, addressZip: 0)
    }
    
    struct StripeResponse: Codable {
        struct Error: LocalizedError, Codable {
            let type: String
            let param: String?
            let message: String
            let code: String?
            
            var errorDescription: String? {
                message
            }
        }
        
        let error: Error?
        let id: String?
    }
    
    func addCard(
        _ card: Card,
        completionHandler: @escaping (Result<Void, SageManager.Error>) -> Void,
        stripeToken: String? = nil
    ) {
        guard let stripeToken = stripeToken else {
            sageManager.getStrikeToken { [weak self] result in
                guard let self = self else { return }
                switch result {
                case .success(let stripeToken):
                    self.addCard(card,
                                 completionHandler: completionHandler,
                                 stripeToken: stripeToken)
                    
                case .failure(let error):
                    completionHandler(.failure(.unexpected(error as NSError)))
                }
            }
            return
        }
        var request = URLRequest(url: URL(string: "https://api.stripe.com/v1/tokens")!)
        request.httpMethod = "POST"
        
        let loginData = "\(stripeToken):".data(using: .utf8)!.base64EncodedString()
        
        request.addValue("Basic \(loginData)", forHTTPHeaderField: "Authorization")
        request.addValue("application/x-www-form-urlencoded; charset=utf-8", forHTTPHeaderField: "Content-Type")

        request.httpBody = card
            .stripeRequestParameters
            .queryParameters
            .data(using: .utf8, allowLossyConversion: true)
        
        URLSession.shared.dataTask(with: request) { [weak self] data, _, error in
            do {
                if let error = error { throw error }
                guard let data = data else { throw NSError(description: "no data") }
                let response = try JSONDecoder().decode(StripeResponse.self, from: data)
                guard let token = response.id else {
                    if let error = response.error {
                        throw error
                    }
                    throw NSError(description: "stripe invalid response")
                }
                self?.sageManager.addPaymentMethod(stripeToken: token, completionHandler: completionHandler)
            } catch {
                DispatchQueue.main.async {
                    completionHandler(.failure(.unexpected(error as NSError)))
                }
            }
        }.resume()
    }
}

extension ConnectionManager: TunnelsManagerActivationDelegate {
    func tunnelActivationAttemptFailed(tunnel: TunnelContainer, error: TunnelsManagerActivationAttemptError) {
        
    }
    
    func tunnelActivationAttemptSucceeded(tunnel: TunnelContainer) {
        
    }
    
    func tunnelActivationFailed(tunnel: TunnelContainer, error: TunnelsManagerActivationError) {
        self.error = error
        if tunnelContainer == nil {
            status = .inactive
        }
    }
    
    func tunnelActivationSucceeded(tunnel: TunnelContainer) {
    }
}

extension ConnectionManager: TunnelsManagerListDelegate {
    func tunnelAdded(at index: Int) {
        tunnelContainer = tunnelsManager.tunnel(named: Self.tunnelName)
    }
    
    func tunnelModified(at index: Int) {
        tunnelContainer = tunnelsManager.tunnel(named: Self.tunnelName)
    }
    
    func tunnelMoved(from oldIndex: Int, to newIndex: Int) {
    }
    
    func tunnelRemoved(at index: Int, tunnel: TunnelContainer) {
        tunnelContainer = tunnelsManager.tunnel(named: Self.tunnelName)
    }
}

extension Data {
    func createSecCertificate() -> SecCertificate? {
        SecCertificateCreateWithData(nil, self as CFData)
    }
}
