// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
import SwiftyUserDefaults
#if os(iOS)
import UIKit
#elseif os(macOS)
#endif

class BubbleManager: Codable, DefaultsSerializable {
    enum URI: NetworkManagerURI {
        case login
        case me
        case device(deviceID: String)
        case devices
        case caCert
        case vpnConf(deviceNameOrUUID: String)
        
        var path: String {
            switch self {
            case .login:
                return "auth/login"
            case .me:
                return "me"
            case .device(let deviceID):
                return "me/devices/\(deviceID)"
            case .devices:
                return "me/devices"
            case .caCert:
                return "auth/cacert"
            case .vpnConf(let deviceNameOrUUID):
                return "me/devices/\(deviceNameOrUUID)/vpn/vpn.conf"
            }
        }
    }
    private enum DeviceType: String, Codable {
        case ios
        case macosx
        
        static let current: Self = {
            #if os(iOS)
            return .ios
            #elseif os(macOS)
            return .macosx
            #endif
        }()
    }
    
    struct RemoteDevice: Codable {
        struct Status: Codable {
            let vpnIps: [String]
        }
        
        let uuid: String
        let name: String
        let status: Status?
    }
    
    private var deviceID: String? {
        didSet {
            BubbleKeychain.shared.currentBubbleManager = self
        }
    }
    private var deviceName: String {
        #if os(iOS)
        let model = UIDevice.current.model
        let name = UIDevice.current.name
        #elseif os(macOS)
        func sysctl(named name: String) -> String {
            var size = 0
            sysctlbyname(name, nil, &size, nil, 0)
            var chars = [CChar](repeating: 0, count: size)
            sysctlbyname(name, &chars, &size, nil, 0)
            return String(cString: chars)
        }
        let model = sysctl(named: "hw.model")
        let name = Host.current().localizedName ?? ProcessInfo.processInfo.fullUserName
        #endif
        return "\(model): \(name)"
    }
    
    typealias NetworkManager = Bubble.NetworkManager<URI>
    typealias Error = NetworkManager.Error
    
    private let networkManager: NetworkManager
    
    init(url: URL) {
        self.networkManager = .init(baseURL: url, appendApiPath: true, token: nil)
    }
    
    // MARK: Public requests
    
    func login(credentials: Credentials,
               completionHandler: @escaping (Result<Void, Error>) -> Void) {
        var request = networkManager.buildRequest(uri: .login, httpMethod: .post)
        request.httpBodyJSON = [
            "name": credentials.username,
            "password": credentials.password,
        ]
        networkManager.perform(request) { result in
            completionHandler(result.flatMap { data in
                guard
                    let dict = try? JSONSerialization.jsonObject(with: data) as? [String: Any],
                    let token = dict["token"] as? String
                    else {
                        return .failure(.invalidResponse)
                }
                self.networkManager.token = token
                return .success(())
            } .mapError {
                switch $0 {
                case .invalidStatusCode(let code, _) where code == 404:
                    return .loginInvalidCredentials
                    
                default:
                    return $0
                }
            })
        }
    }
    
    func validateSSL(invalidSSLHandler: @escaping () -> Void) -> Cancellable {
        var cancellableContainer: CancellableContainer!
        let task = URLSession.shared.dataTask(with: .init(url: URL(string: "https://example.com/")!)) { _, _, error in
            guard let error = error as NSError? else {
                wg_log(.debug, message: "ssl is valid")
                return
            }
            switch error.code {
            case NSURLErrorClientCertificateRequired...NSURLErrorSecureConnectionFailed:
                DispatchQueue.main.async {
                    invalidSSLHandler()
                }
                
            case NSURLErrorTimedOut:
                DispatchQueue.main.asyncAfter(deadline: .now() + 10) { [weak self] in
                    guard let self = self else { return }
                    cancellableContainer.cancellable = self.validateSSL(invalidSSLHandler: invalidSSLHandler)
                }
                
            default:
                wg_log(error: error, message: "\(Self.self) \(#function) non ssl error")
            }
        }
        cancellableContainer = CancellableContainer(cancellable: task)
        task.resume()
        return cancellableContainer
    }
    
    var myBubbleURL: URL? {
        networkManager.buildBrowserURL(uri: nil)
    }
    var accountURL: URL? {
        networkManager.buildBrowserURL(uri: .me)
    }
    var sessionToken: String? {
        networkManager.token
    }
    var host: String {
        networkManager.baseURL.host!
    }
    
    var caCertURL: URL {
        networkManager.buildURL(
            uri: .caCert,
            urlParams: ["deviceType": DeviceType.current.rawValue,
                        "type": "cer"])
    }
    #if os(macOS)
    func getCACertificate(completionHandler: @escaping (Result<URL, Error>) -> Void) {
        let url = URL(fileURLWithPath: NSTemporaryDirectory()).appendingPathComponent("ca-cert")
        networkManager.performAndWrite(
            .init(url: caCertURL,
                  httpMethod: .get),
            outputURL: url,
            completionHandler: completionHandler)
        
    }
    #endif
    
    func getVPNConf(completionHandler: @escaping (Result<URL, Error>) -> Void) -> Cancellable {
        var cancellableContainer: CancellableContainer!
        let cancellable = addDeviceIfNeeded { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .failure(let error):
                completionHandler(.failure(error))
                
            case .success(let deviceUUID):
                let url = URL(fileURLWithPath: NSTemporaryDirectory()).appendingPathComponent("vpn.conf")
                let request = self.networkManager.buildRequest(
                    uri: .vpnConf(deviceNameOrUUID: deviceUUID),
                    httpMethod: .get
                )
                cancellableContainer.cancellable = self.networkManager.performAndWrite(
                    request,
                    outputURL: url,
                    completionHandler: completionHandler
                )
            }
        }
        cancellableContainer = .init(cancellable: cancellable)
        return cancellableContainer
    }
    
    // MARK: Private requests
    
    private func addDeviceIfNeeded(completionHandler: @escaping (Result<String, Error>) -> Void) -> Cancellable {
        var cancellableContainer: CancellableContainer!
        let cancellable = listDevices { [weak self] result in
            guard let self = self else { return }
            switch result {
            case .failure(let error):
                completionHandler(.failure(error))
                
            case .success(let devices):
                if let deviceID = self.deviceID,
                    (devices.contains { $0.uuid == deviceID })
                {
                    completionHandler(.success(deviceID))
                    return
                }
                let originalDeviceName = self.deviceName
                var newDeviceName = originalDeviceName
                var i = 1
                while (devices.contains { $0.name == newDeviceName}) {
                    newDeviceName = "\(originalDeviceName) \(i)"
                    i += 1
                }
                cancellableContainer.cancellable = self.addDevice(name: newDeviceName) { [weak self] result in
                    guard let self = self else { return }
                    completionHandler(result.map { device in
                        self.deviceID = device.uuid
                        return device.uuid
                    })
                }
            }
        }
        cancellableContainer = .init(cancellable: cancellable)
        return cancellableContainer
    }
    
    private func listDevices(completionHandler: @escaping (Result<[RemoteDevice], Error>) -> Void) -> Cancellable {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .devices,
                httpMethod: .get,
                cachePolicy: .reloadIgnoringLocalAndRemoteCacheData
            ))
        { result in
            try result.processAndRedirect(
                process: { [weak self] data in
                    let devices = try JSONDecoder().decode([RemoteDevice].self, from: data)
                    self?.removeOld(from: devices)
                    return devices
                },
                completionHandler: completionHandler)
        }
    }
    
    private func removeOld(from devices: [RemoteDevice]) {
        Defaults.deviceIDsToBeDeleted = Defaults.deviceIDsToBeDeleted.filter { deviceIDToBeDeleted in
            guard
                deviceIDToBeDeleted != deviceID,
                (devices.contains { $0.uuid == deviceIDToBeDeleted })
                else { return false }
            removeDevice(deviceID: deviceIDToBeDeleted)
            return true
        }
    }
    
    private func addDevice(
        name: String,
        completionHandler: @escaping (Result<RemoteDevice, Error>) -> Void)
        -> Cancellable
    {
        var request = networkManager.buildRequest(uri: .devices, httpMethod: .put)
        request.httpBodyJSON = [
            "name": name,
            "deviceType": DeviceType.current.rawValue,
        ]
        return networkManager.perform(request) { result in
            try result.processAndRedirect(
                process: { data in
                    let device = try JSONDecoder().decode(RemoteDevice.self, from: data)
                    Defaults.deviceIDsToBeDeleted.append(device.uuid)
                    return device
            },
                completionHandler: completionHandler
            )
        }
    }
    
    func getCurrentDevice(
        completionHandler: @escaping (Result<RemoteDevice, Error>) -> Void)
    -> Cancellable?
    {
        guard let deviceID = deviceID else { return nil }
        return networkManager.perform(
            networkManager.buildRequest(
                uri: .device(deviceID: deviceID),
                httpMethod: .get
            )
        ) { result in
            try result.processAndRedirect(
                process: { try JSONDecoder().decode(RemoteDevice.self, from: $0) },
                completionHandler: completionHandler
            )
        }
    }
    
    private func removeDevice(deviceID: String) {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .device(deviceID: deviceID),
                httpMethod: .delete
        ))
    }
        
    // MARK: Coding
    
    enum CodingKeys: String, CodingKey {
        case url
        case token
        case deviceID
    }
    
    required init(from decoder: Decoder) throws {
        let values = try decoder.container(keyedBy: CodingKeys.self)
        let baseURL = try values.decode(URL.self, forKey: .url)
        let token = try values.decode(String.self, forKey: .token)
        deviceID = try? values.decode(String.self, forKey: .deviceID)
        networkManager = .init(baseURL: baseURL, appendApiPath: true, token: token)
        wg_log(.info, message: "\(Self.self) \(#function) \(baseURL) \(token)")
    }
    
    func encode(to encoder: Encoder) throws {
        guard let token = networkManager.token else { throw NSError(description: "") }
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(networkManager.baseURL, forKey: .url)
        try container.encode(token, forKey: .token)
        try container.encode(deviceID, forKey: .deviceID)
    }
    
    func destroy() {
        if let deviceID = deviceID {
            removeDevice(deviceID: deviceID)
        }
        networkManager.destroy()
    }
}
