// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation

class FlexrouterManager {
    enum URI: String, NetworkManagerURI {
        case ping
        case register
        case unregister
        
        var path: String { rawValue }
    }
    struct Info: Codable {
        let password: String
        let session: String
        let bubble: String
        let ip: String
    }
    enum Error: LocalizedError {
        case invalidPassword
        
        case networkError(NetworkManager.Error)
        
        static func error(with error: NetworkManager.Error) -> Self {
            if case .invalidStatusCode(let code, let data) = error,
               code == 401,
               data?.utf8String?.contains("password was incorrect") == true
            {
                return .invalidPassword
            }
            return .networkError(error)
        }
        
        var errorDescription: String? {
            switch self {
            case .invalidPassword:
                return "Invalid password"
            case .networkError(let error):
                return error.localizedDescription
            }
        }
    }
    
    typealias NetworkManager = Bubble.NetworkManager<URI>
    
    private let networkManager: NetworkManager
    
    init() {
        networkManager = .init(
            baseURL: URL(string: "http://127.0.0.1:9833/")!,
            appendApiPath: false,
            token: nil
        )
    }
    
    @discardableResult
    func ping(completionHandler: @escaping (Result<Void, Error>) -> Void)
    -> Cancellable
    {
        var request = networkManager.buildRequest(uri: .ping, httpMethod: .get)
        request.timeoutInterval = 1
        return networkManager.perform(request) {
            completionHandler(
                $0.map { _ in }
                    .mapError(Error.error)
            )
        }
    }
    
    @discardableResult
    func register(_ info: Info, completionHandler: @escaping (Result<Void, Error>) -> Void)
    -> Cancellable
    {
        var request = networkManager.buildRequest(uri: .register, httpMethod: .post)
        request.httpBody = try? JSONEncoder().encode(info)
        return networkManager.perform(request) {
            completionHandler(
                $0.map { _ in }
                    .mapError(Error.error)
            )
        }
    }
    
    @discardableResult
    func unregister(
        password: String,
        completionHandler: @escaping (Result<Void, Error>) -> Void
    ) -> Cancellable
    {
        var request = networkManager.buildRequest(uri: .unregister, httpMethod: .post)
        request.httpBodyJSON = [
            "password": password
        ]
        return networkManager.perform(request) {
            completionHandler(
                $0.map { _ in }
                    .mapError(Error.error)
            )
        }
    }
}
