// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
#if os(iOS)
import UIKit
#elseif os(macOS)
#endif

protocol NetworkManagerURI {
    var path: String { get }
}
struct NetworkManagerServerError: Codable, Equatable, CustomStringConvertible {
    let messageTemplate: String
    let message: String?
    
    var description: String {
        message ?? messageTemplate
    }
}

enum NetworkManagerError: LocalizedError, Equatable {
    case tokenExpired
    case networkError(Int)
    case loginInvalidCredentials
    case invalidResponse
    case invalidStatusCode(Int, Data?)
    
    case unexpected(NSError)
    case serverErrors([NetworkManagerServerError])
    
    var errorDescription: String? {
        switch self {
        case .tokenExpired:
            return "Token expired"
        case .networkError:
            return "The Bubble is unavailable"
        case .loginInvalidCredentials:
            return "Invalid username/password specified"
        case .invalidResponse:
            return "There was an unexpected error"
        case .invalidStatusCode(let statusCode, let data):
            return "Invalid status code: \(statusCode) \(data?.utf8String ?? "")"
        case .unexpected:
            return "There was a network error communicating with this Bubble"
        case .serverErrors(let serverErrors):
            return serverErrors.reduce(into: "") {
                $0 += "\($1.description)\n"
            }
        }
    }
}

enum HTTPMethod: String {
    case post = "POST"
    case get = "GET"
    case put = "PUT"
    case delete = "DELETE"
}

class NetworkManager<URI: NetworkManagerURI> {
    typealias ServerError = NetworkManagerServerError
    typealias Error = NetworkManagerError
    
    let baseURL: URL
    private let appendApiPath: Bool
    private var apiURL: URL {
        if appendApiPath {
            return baseURL.appendingPathComponent("api")
        }
        return baseURL
    }
    
    init(baseURL: URL, appendApiPath: Bool, token: String?) {
        self.appendApiPath = appendApiPath
        self.baseURL = baseURL
        self.token = token
        urlSession = Self.createURLSessionConfiguration(token: token)
    }
    
    var token: String? {
        didSet {
            urlSession = Self.createURLSessionConfiguration(token: token)
        }
    }
    private var urlSession: URLSession {
        didSet {
            oldValue.invalidateAndCancel()
        }
    }
    
    @discardableResult func performAndWrite(
        _ request: URLRequest,
        outputURL: URL,
        completionHandler: @escaping (Result<URL, Error>) throws -> Void)
        -> Cancellable
    {
        perform(request) { result in
            try result.processAndRedirect(
                process: { data in
                    try? FileManager.default.removeItem(at: outputURL)
                    try data.write(to: outputURL)
                    return outputURL
            },
                completionHandler: completionHandler
            )
        }
    }
    
    @discardableResult func perform(
        _ request: URLRequest,
        completionHandler: ((Result<Data, Error>) throws -> Void)? = nil)
        -> Cancellable
    {
        var task: URLSessionDataTask!
        networkProcessing = true
        task = urlSession.dataTask(with: request) { [weak self] data, response, error in
            self?.networkProcessing = false
            guard let completionHandler = completionHandler else { return }
            func logResponseError(_ appender: String? = nil) {
                let info: [String: CustomStringConvertible?] = [
                    "currentRequest": task.currentRequest,
                    "httpData": task.currentRequest?.httpBodyJSON,
                    "method": task.currentRequest?.httpMethod,
                    "response": response,
                    "appender": appender,
                    "data": data?.encodedString(.utf8),
                ]
                wg_log(.info, message: info.map { "\($0): \($1?.description as AnyObject)" }.joined(separator: "; "))
            }
            do {
                if let error = error as NSError? {
                    let failure: Error
                    if error.domain == NSURLErrorDomain {
                        failure = .networkError(error.code)
                    } else {
                        failure = .unexpected(error)
                    }
                    try? completionHandler(.failure(failure))
                    return
                }
                guard
                    let data = data,
                    let httpResponse = response as? HTTPURLResponse else {
                        logResponseError()
                        try? completionHandler(.failure(.invalidResponse))
                        return
                }
                switch httpResponse.statusCode {
                case (200...299):
                    try completionHandler(.success(data))
                    
                case 422:
                    do {
                        let errors = try JSONDecoder().decode([ServerError].self, from: data)
                        logResponseError()
                        try? completionHandler(.failure(.serverErrors(errors)))
                    } catch {
                        logResponseError("parse 422 error: \(error)")
                        try? completionHandler(.failure(.invalidStatusCode(httpResponse.statusCode, data)))
                    }
                    
                case 403 where self?.token != nil:
                    try? completionHandler(.failure(.tokenExpired))
                    
                default:
                    try completionHandler(.failure(.invalidStatusCode(httpResponse.statusCode, data)))
                }
            } catch {
                logResponseError("\(error)")
            }
        }
        task.resume()
        return task
    }
    
    private var networkProcessing: Bool = false {
        didSet {
            #if os(iOS)
            DispatchQueue.main.async {
                UIApplication.shared.isNetworkActivityIndicatorVisible = self.networkProcessing
            }
            #elseif os(macOS)
            #endif
        }
    }
    
    func buildRequest(
        uri: URI,
        urlParams: [String: String]? = nil,
        httpMethod: HTTPMethod,
        cachePolicy: URLRequest.CachePolicy = .useProtocolCachePolicy) -> URLRequest
    {
        .init(
            url: buildURL(uri: uri, urlParams: urlParams),
            httpMethod: httpMethod,
            cachePolicy: cachePolicy
        )
    }
    
    func buildURL(uri: URI, urlParams: [String: String]? = nil) -> URL {
        var url = apiURL.appendingPathComponent(uri.path)
        if let urlParams = urlParams {
            url.appendQueryParameters(urlParams)
        }
        return url
    }
    
    func buildBrowserURL(uri: URI?) -> URL? {
        guard let token = token else { return nil }
        var path = uri?.path ?? ""
        if path.first != "/" {
            path = "/" + path
        }
        return baseURL
            .appendingPathComponent("appLogin")
            .appendingQueryParameters(
                ["session": token,
                 "uri": path]
        )
    }
    
    private static func createURLSessionConfiguration(token: String?) -> URLSession {
        let configuration = URLSessionConfiguration.default
        if let token = token {
            configuration.urlCache = URLCache(
                memoryCapacity: 512_000,
                diskCapacity: 10_000_000,
                diskPath: "\(token)")
        } else {
            configuration.urlCache = nil
        }
        configuration.requestCachePolicy = .returnCacheDataElseLoad
        configuration.httpAdditionalHeaders = [
            "Content-Type": "application/json",
        ]
        configuration.httpAdditionalHeaders!["X-Bubble-Session"] = token
        return .init(configuration: configuration, delegate: nil, delegateQueue: .main)
    }
    
    func destroy() {
        urlSession.finishTasksAndInvalidate()
        urlSession.configuration.urlCache?.removeAllCachedResponses()
        networkProcessing = false
    }
}

extension URLRequest {
    var httpBodyJSON: [String: Any] {
        get {
            guard
                let httpBody = httpBody,
                let result = try? JSONSerialization.jsonObject(with: httpBody) as? [String: Any]
                else { return [:] }
            return result
        }
        set {
            httpBody = try? JSONSerialization.data(withJSONObject: newValue)
        }
    }
    
    var description: String {
        "\(self) method:\(httpMethod ?? "none") body: \(httpBodyJSON)"
    }
}

private extension Data {
    func encodedString(_ encoding: String.Encoding) -> String? {
        String(data: self, encoding: encoding)
    }
}

extension Array where Element == NetworkManagerServerError {
    var description: String {
        reduce(into: "") {
            $0 += $1.description
        }
    }
}

private protocol URLQueryParameterStringConvertible {
    var queryParameters: String {get}
}

extension Dictionary: URLQueryParameterStringConvertible {
    /**
     This computed property returns a query parameters string from the given NSDictionary. For
     example, if the input is @{@"day":@"Tuesday", @"month":@"January"}, the output
     string will be @"day=Tuesday&month=January".
     @return The computed parameters string.
     */
    var queryParameters: String {
        compactMap { key, value in
            return String(format: "%@=%@",
                          String(describing: key).addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!,
                          String(describing: value).addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!)
        }.joined(separator: "&")
    }
}

extension URL {
    /**
     Creates a new URL by adding the given query parameters.
     @param parametersDictionary The query parameter dictionary to add.
     @return A new URL.
     */
    mutating func appendQueryParameters(_ parametersDictionary: [String: String]) {
        self = self.appendingQueryParameters(parametersDictionary)
    }
    
    func appendingQueryParameters(_ parametersDictionary: [String: String]) -> URL {
        URL(string: String(format: "%@?%@", self.absoluteString, parametersDictionary.queryParameters))!
    }
}

extension Result where Success == Data, Failure == NetworkManagerError {
    func processAndRedirect<Success>(
        process: (Data) throws -> (Success),
        completionHandler: @escaping (Result<Success, Failure>) throws -> Void)
        throws
    {
        var insideMapError: Swift.Error?
        try completionHandler(flatMap { data in
            do {
                let success = try process(data)
                return .success(success)
            } catch {
                insideMapError = error
                return .failure(.invalidResponse)
            }
        })
        if let insideMapError = insideMapError {
            throw insideMapError
        }
    }
}

extension URLRequest {
    init(url: URL,
         httpMethod: HTTPMethod,
         cachePolicy: CachePolicy = .useProtocolCachePolicy)
    {
        self.init(url: url)
        self.httpMethod = httpMethod.rawValue
        self.cachePolicy = cachePolicy
    }
}
