// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation

class SageFabric {
    private struct Bootstrap: Decodable {
        let sages: [URL]
        
        static let zero = Self(sages: [])
    }
    
    func login(
        credentials: Credentials,
        customURL: URL?,
        completionHandler: @escaping (Result<SageManager, SageManager.Error>) -> Void
    ) {
        if let customURL = customURL {
            tryLogin(
                credentials: credentials,
                urls: [customURL],
                index: 0,
                completionHandler: completionHandler
            )
            return
        }
        getBootstrapURLs { [weak self] result in
            switch result {
            case .failure(let error):
                completionHandler(.failure(.networkError(.unexpected(error as NSError))))
                
            case .success(let urls):
                guard !urls.isEmpty else {
                    completionHandler(.failure(.networkError(.invalidResponse)))
                    return
                }
                self?.tryLogin(
                    credentials: credentials,
                    urls: urls,
                    index: 0,
                    completionHandler: completionHandler
                )
            }
        }
    }
    
    func getForgotPasswordURL(completionHandler: @escaping (Result<URL, Error>) -> Void) {
        getBootstrapURLs { result in
            switch result {
            case .failure(let error):
                completionHandler(.failure(error))
                
            case .success(let urls):
                guard let firstSage = urls.first else {
                    completionHandler(.failure(NSError(description: "empty sage")))
                    return
                }
                completionHandler(.success(firstSage.appendingPathComponent("forgotPassword")))
            }
        }
    }
    
    func getRegisterURL(completionHandler: @escaping (Result<URL, Error>) -> Void) {
        getBootstrapURLs { result in
            switch result {
            case .failure(let error):
                completionHandler(.failure(error))
                
            case .success(let urls):
                guard let firstSage = urls.first else {
                    completionHandler(.failure(NSError(description: "empty sage")))
                    return
                }
                completionHandler(.success(firstSage.appendingPathComponent("register")))
            }
        }
    }
    
    private func tryLogin(
        credentials: Credentials,
        urls: [URL],
        index: Int,
        completionHandler: @escaping (Result<SageManager, SageManager.Error>) -> Void
    ) {
        guard index < urls.count else {
            completionHandler(.failure(.networkError(.loginInvalidCredentials)))
            return
        }
        let sageManager = SageManager(sageURL: urls[index])
        let date = Date()
        sageManager.login(credentials: credentials) { result in
            switch result {
            case .failure(let error):
                if error == .networkError(.loginInvalidCredentials) {
                    self.tryLogin(
                        credentials: credentials,
                        urls: urls,
                        index: index + 1,
                        completionHandler: completionHandler
                    )
                } else {
                    completionHandler(.failure(error))
                }
                
            case .success(let manager):
                completionHandler(.success(manager))
            }
            wg_log(.debug, message: "\(#function), \(-date.timeIntervalSinceNow), \(urls[index]), \(index), \(result)")
        }
    }
    
    private func getBootstrapURLs(completionHandler: @escaping (Result<[URL], Error>) -> Void) {
        let request = URLRequest(
            url: URL(string: "https://raw.githubusercontent.com/getbubblenow/bubble-config/master/boot.json")!,
            cachePolicy: .reloadRevalidatingCacheData)
        URLSession.shared.dataTask(with: request) { data, _, error in
            do {
                guard error == nil else { throw error! }
                guard let data = data else { throw NSError(description: "no data") }
                let result = try JSONDecoder().decode(Bootstrap.self, from: data).sages
                guard !result.isEmpty else { throw NSError(description: "No sage urls") }
                completionHandler(.success(result))
            } catch {
                completionHandler(.failure(error))
            }
        }.resume()
    }
}
