// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
import SwiftyUserDefaults

class SageManager: Codable, DefaultsSerializable {
    typealias NetworkManager = Bubble.NetworkManager<URI>
    typealias NetworkError = NetworkManager.Error
    
    enum Error: LocalizedError, Equatable {
        case noBubble
        case emptyCredentials
        case invalidResponse(info: String)
        case networkError(NetworkError)
        case unexpected(NSError)
        
        var errorDescription: String? {
            switch self {
            case .noBubble:
                return "No Bubble"
            case .emptyCredentials:
                return "Empty credentials"
            case .invalidResponse(let info):
                return "Invalid response: \(info)"
            case .networkError(let networkError):
                return networkError.errorDescription
            case .unexpected(let error):
                return error.fixedLocalizedDescription
            }
        }
    }
    
    enum URI: NetworkManagerURI {
        case login
        case newBubble
        case supportedPaymentMethods
        case me
        case networks
        case plans
        case payment
        case paymentMethods
        case planActionStart(planName: String)
        case networkStatus(userId: String, networkId: String)
        
        var path: String {
            switch self {
            case .login:
                return "auth/login"
            case .newBubble:
                return "new_bubble"
            case .supportedPaymentMethods:
                return "paymentMethods"
            case .me:
                return "me"
            case .networks:
                return "me/networks"
            case .payment:
                return "me/payment"
            case .paymentMethods:
                return "me/paymentMethods"
            case .plans:
                return "me/plans"
            case .planActionStart(let planName):
                return "me/networks/\(planName)/actions/start"
            case .networkStatus(let userId, let networkId):
                return "users/\(userId)/networks/\(networkId)/actions/status"
            }
        }
    }
    struct Network: Codable {
        enum State: String, CodableLastDefault {
            case running
            case created
            case starting
            case restoring
            
            case unknown
            
            var current: Bool {
                switch self {
                case .running, .starting, .restoring:
                    return true
                    
                default:
                    return false
                }
            }
        }
        struct Status: Codable, Equatable, CustomStringConvertible {
            let messageKey: String
            var percent: Double { _percent ?? 0 }
            private let _percent: Double?
            
            enum CodingKeys: String, CodingKey {
                case _percent = "percent" // swiftlint:disable:this identifier_name
                case messageKey
            }
            
            var description: String { "Status(messageKey: \"\(messageKey)\", percent: \(percent))" }
        }
        
        let uuid: String
        let name: String
        let domainName: String
        let state: State
        
        var nodeBaseURL: URL? { URL(sslSafe: "\(name).\(domainName)", forcePort: 1443) }
    }
    private struct AccountPaymentMethod: Codable {
        enum PaymentMethodType: String, CodableLastDefault {
            case credit
            case promotion = "promotional_credit"
            
            case unknown
        }
        
        let paymentMethodType: PaymentMethodType
        let paymentInfo: String?
    }
    struct AccountPlan: Codable {
        let name: String
    }
    
    private let networkManager: NetworkManager
    private var credentials: Credentials?
    
    var myBubbleURL: URL? { networkManager.buildBrowserURL(uri: nil) }
    var accountURL: URL? { networkManager.buildBrowserURL(uri: .me) }
    var launchBubbleURL: URL? { networkManager.buildBrowserURL(uri: .newBubble) }
    var paymentPageURL: URL? { networkManager.buildBrowserURL(uri: .payment) }
    
    init(sageURL: URL) {
        networkManager = .init(baseURL: sageURL, appendApiPath: true, token: nil)
    }
    
    func login(
        credentials: Credentials,
        completionHandler: @escaping (Result<SageManager, Error>) -> Void
    ) {
        var request = networkManager.buildRequest(uri: .login, httpMethod: .post)
        request.httpBodyJSON = [
            "name": credentials.username,
            "password": credentials.password,
        ]
        networkManager.perform(request) { result in
            completionHandler(result.flatMap { data in
                guard
                    let dict = try? JSONSerialization.jsonObject(with: data) as? [String: Any],
                    let token = dict["token"] as? String
                    else {
                        return .failure(.invalidResponse)
                }
                self.networkManager.token = token
                self.credentials = credentials
                return .success(self)
            } .mapError {
                switch $0 {
                case .invalidStatusCode(let code, _) where code == 404:
                    return .networkError(.loginInvalidCredentials)
                    
                default:
                    return .networkError($0)
                }
            })
        }
    }
    
    struct BubbleManagerGetterProps {
        typealias StatusHandler = (Network.Status) -> Void
        typealias CompletionHandler = (Result<BubbleManager, Error>) -> Void
        
        let bubbleStartingStatusHandler: StatusHandler
        let completionHandler: CompletionHandler
        fileprivate var attempt: Int = 0
        
        init(bubbleStartingStatusHandler: @escaping StatusHandler, completionHandler: @escaping CompletionHandler) {
            self.bubbleStartingStatusHandler = bubbleStartingStatusHandler
            self.completionHandler = completionHandler
        }
    }
    
    func getBubbleManager(_ props: BubbleManagerGetterProps) {
        guard let credentials = credentials else {
            props.completionHandler(.failure(.emptyCredentials))
            return
        }
        networkManager.perform(
            networkManager.buildRequest(uri: .networks, httpMethod: .get)
        ) { result in
            switch result {
            case .success(let data):
                let networks = try JSONDecoder().decode([Network].self, from: data)
                guard let network = (networks.first { $0.state.current })
                    else {
                        props.completionHandler(.failure(.noBubble))
                        return
                }
                switch network.state {
                case .running:
                    let bubbleManager = BubbleManager(url: network.nodeBaseURL!)
                    bubbleManager.login(credentials: credentials) { result in
                        if case .failure(let error) = result,
                            case .serverErrors = error,
                            props.attempt < 3
                        {
                            wg_log(.debug, #function, error, "retrying")
                            var props = props
                            props.attempt += 1
                            DispatchQueue.main.asyncAfter(deadline: .now() + 3) {
                                self.getBubbleManager(props)
                            }
                        } else {
                            props.completionHandler(
                                result.map { _ in bubbleManager }
                                    .mapError(self.mapNetworkError)
                            )
                        }
                    }
                    
                default:
                    // waiting until bubble is up and try again
                    self.updateStateUntilRunning(
                        network: network,
                        statusHandler: props.bubbleStartingStatusHandler
                    ) { [weak self] result in
                        switch result {
                        case .success:
                            self?.getBubbleManager(props)
                            
                        case .failure(let error):
                            props.completionHandler(.failure(error))
                        }
                    }
                }
                
            case .failure(let error):
                props.completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    private func updateStateUntilRunning(
        network: Network,
        statusHandler: @escaping BubbleManagerGetterProps.StatusHandler,
        prevStatus: Network.Status? = nil,
        completionHandler: @escaping (Result<Void, Error>) -> Void
    ) {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .networkStatus(userId: credentials!.username, networkId: network.uuid),
                httpMethod: .get
            )
        ) { result in
            switch result {
            case .success(let data):
                let statuses = try JSONDecoder().decode([Network.Status].self, from: data)
                guard let status = statuses.first else {
                    completionHandler(.failure(.noBubble))
                    return
                }
                
                guard status.percent < 100,
                    prevStatus == nil || prevStatus!.percent <= status.percent
                    else {
                        completionHandler(.success(()))
                        return
                }
                
                wg_log(.info, message: "\(Self.self) \(#function) \(status.percent) \(status.messageKey)")
                if prevStatus != status {
                    statusHandler(status)
                }
                DispatchQueue.main.asyncAfter(deadline: .now() + 5) { [weak self] in
                    self?.updateStateUntilRunning(
                        network: network,
                        statusHandler: statusHandler,
                        prevStatus: status,
                        completionHandler: completionHandler
                    )
                }
                
            case .failure(let error):
                completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    func hasValidPaymentMethod(completionHandler: @escaping (Result<Bool, Error>) -> Void) {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .paymentMethods,
                httpMethod: .get
            )
        ) { result in
            switch result {
            case .success(let data):
                let paymentMethods = try JSONDecoder().decode([AccountPaymentMethod].self, from: data)
                if (paymentMethods.first { $0.paymentMethodType != .promotion }) != nil {
                    completionHandler(.success(true))
                } else {
                    completionHandler(.success(false))
                }

            case .failure(let error):
                completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    func addPaymentMethod(stripeToken: String, completionHandler: @escaping (Result<Void, Error>) -> Void) {
        var request = networkManager.buildRequest(uri: .paymentMethods, httpMethod: .put)
        do {
            request.httpBody = try JSONEncoder().encode(
                AccountPaymentMethod(paymentMethodType: .credit, paymentInfo: stripeToken)
            )
            networkManager.perform(request) { result in
                switch result {
                case .success:
                    completionHandler(.success(()))
                    
                case .failure(let error):
                    completionHandler(.failure(.networkError(error)))
                }
            }
        } catch {
            completionHandler(.failure(.networkError(.unexpected(error as NSError))))
        }
    }
    
    func createAccountPlan(completionHandler: @escaping (Result<AccountPlan, Error>) -> Void) {
        var request = networkManager.buildRequest(uri: .plans, httpMethod: .put)
        request.httpBodyJSON = [:]
        networkManager.perform(request) { result in
            switch result {
            case .success(let data):
                completionHandler(.success(try JSONDecoder().decode(AccountPlan.self, from: data)))
                
            case .failure(let error):
                completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    func launchBubble(_ accountPlan: AccountPlan, completionHandler: @escaping (Result<Void, Error>) -> Void) {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .planActionStart(planName: accountPlan.name),
                httpMethod: .post
            )
        ) { result in
            switch result {
            case .success:
                completionHandler(.success(()))
                
            case .failure(let error):
                completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    func getStrikeToken(completionHandler: @escaping (Result<String, Error>) -> Void) {
        networkManager.perform(
            networkManager.buildRequest(
                uri: .supportedPaymentMethods,
                httpMethod: .get
            )
        ) { result in
            switch result {
            case .success(let data):
                let json = try JSONSerialization.jsonObject(with: data)
                guard
                    let stripeJson = ((json as? [[String: Any]])?.first {
                        $0["driverClass"] as? String == "bubble.cloud.payment.stripe.StripePaymentDriver"
                    }),
                    let stripeToken = (stripeJson["driverConfig"] as? [String: Any])?["publicApiKey"] as? String,
                    stripeToken.starts(with: "pk_")
                    else {
                        completionHandler(.failure(.invalidResponse(info: "\(json)")))
                        return
                }
                completionHandler(.success(stripeToken))
                
            case .failure(let error):
                completionHandler(.failure(.networkError(error)))
            }
        }
    }
    
    // MARK: Coding
    
    enum CodingKeys: String, CodingKey {
        case url
        case token
        case credentials
    }
    
    required init(from decoder: Decoder) throws {
        let values = try decoder.container(keyedBy: CodingKeys.self)
        let baseURL = try values.decode(URL.self, forKey: .url)
        let token = try values.decode(String.self, forKey: .token)
        networkManager = .init(baseURL: baseURL, appendApiPath: true, token: token)
        credentials = try values.decode(Credentials.self, forKey: .credentials)
        wg_log(.info, message: "\(Self.self) \(#function) \(token) \(credentials!)")
    }
    
    func encode(to encoder: Encoder) throws {
        guard
            let token = networkManager.token,
            let credentials = credentials
            else { throw NSError(description: "no token") }
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(networkManager.baseURL, forKey: .url)
        try container.encode(token, forKey: .token)
        try container.encode(credentials, forKey: .credentials)
    }
    
    func destroy() {
        networkManager.destroy()
    }
    
    private func mapNetworkError(_ networkError: NetworkError) -> Error {
        .networkError(networkError)
    }
}
