// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Lottie

class ConnectionView: View {
    struct Props: Equatable {
        enum State: Equatable {
            case noBubble
            case noPaymentMethod
            case starting(progress: Double)
            case disconnected
            case connecting
            case connected
            
            var isStarting: Bool {
                switch self {
                case .starting:
                    return true
                    
                default:
                    return false
                }
            }
        }
        enum ButtonState {
            case hidden
            case connect
            case connecting
            case disconnect
            case disconnecting
            case restarting
            
            var description: String {
                switch self {
                case .hidden: return ""
                case .connect: return "Connect"
                case .connecting: return "Connecting..."
                case .disconnect: return "Disconnect"
                case .disconnecting: return "Disconnecting..."
                case .restarting: return "Restarting..."
                }
            }
            
            var visible: Bool {
                switch self {
                case .connect, .disconnect:
                    return true
                case .hidden, .connecting, .disconnecting, .restarting:
                    return false
                }
            }
        }
        
        var buttonState: ButtonState
        var state: State
        
        init(status: ConnectionManager.Status) {
            switch status {
            case .initial:
                buttonState = .hidden
                state = .disconnected
            case .noBubble:
                buttonState = .hidden
                state = .noBubble
            case .noPaymentMethod:
                buttonState = .hidden
                state = .noPaymentMethod
            case .startingBubble(let progress):
                buttonState = .hidden
                state = .starting(progress: progress)
            case .inactive:
                buttonState = .connect
                state = .disconnected
            case .activating:
                buttonState = .connecting
                state = .connecting
            case .active:
                buttonState = .disconnect
                state = .connected
            case .deactivating:
                buttonState = .disconnecting
                state = .disconnected
            case .restarting:
                buttonState = .restarting
                state = .connecting
            }
        }
        
        private init(buttonState: ButtonState, state: State) {
            self.buttonState = buttonState
            self.state = state
        }
        
        static let zero = Self(buttonState: .connect, state: .disconnected)
    }
    
    @IBOutlet weak var noBubbleView: View!
    @IBOutlet weak var disconnectedView: View!
    @IBOutlet weak var connectedView: View!
    @IBOutlet weak var connectingView: View!
    @IBOutlet weak var connectButton: Button!
    @IBOutlet weak var noPaymentView: View!
    
    @IBOutlet weak var startingView: View!
    @IBOutlet weak var startingAnimationView: AnimationView!
    @IBOutlet weak var startingProgressLabel: Label!
    
    @IBOutlet weak var noBubbleAnimationView: AnimationView!
    @IBOutlet weak var connectingAnimationView: AnimationView!
    @IBOutlet weak var connectedAnimationView: AnimationView!
    
    var props: Props = .zero {
        didSet {
            guard oldValue != props else { return }
            propsUpdated(oldValue: oldValue)
        }
    }
    
    override func awakeFromNib() {
        super.awakeFromNib()
        startingView.backgroundColor = R.color.connectingBackground()
        noBubbleView.backgroundColor = R.color.noBubbleBackground()
        noBubbleAnimationView.animation = Animation.fileURL(R.file.noBubbleAnimationJson()!)
        startingAnimationView.animation = Animation.fileURL(R.file.bubbleStartingAnimationJson()!)
        connectingAnimationView.animation = Animation.fileURL(R.file.connectingAnimationJson()!)
        connectingAnimationView.loopMode = .loop
        connectedAnimationView.animation = Animation.fileURL(R.file.connectedAnimationJson()!)
        #if os(iOS)
        backgroundColor = .clear
        #elseif os(macOS)
        backgroundColor = .white
        shadow = NSShadow()
        backingLayer.cornerRadius = 10
        startingView.backingLayer.cornerRadius = backingLayer.cornerRadius
        backingLayer.shadowColor = Color(red: 0.067, green: 0.067, blue: 0, alpha: 0.1).cgColor
        backingLayer.shadowOpacity = 1
        backingLayer.shadowRadius = 50
        backingLayer.shadowOffset = CGSize(width: 0, height: 10)
        #endif
        propsUpdated()
    }
    
    private func propsUpdated(oldValue: Props? = nil) {
        connectButton.isHidden = !props.buttonState.visible
        connectButton.title = props.buttonState.description.uppercased()
        guard oldValue?.state != props.state else { return }
        var currentView: View
        var currentAnimationView: AnimationView?
        switch props.state {
        case .noBubble:
            currentView = noBubbleView
            currentAnimationView = noBubbleAnimationView
            
        case .starting(let progress):
            currentView = startingView
            startingProgressLabel.text = "\(Int(progress * 100))%"
            if oldValue?.state.isStarting != true {
                startingAnimationView.currentFrame = 0
            }
            currentAnimationView = startingAnimationView
            startingAnimationView.play(toFrame: connectingAnimationFrame(for: progress))
            
        case .connecting:
            currentView = connectingView
            currentAnimationView = connectingAnimationView
            
        case .disconnected:
            currentView = disconnectedView
            
        case .connected:
            currentView = connectedView
            currentAnimationView = connectedAnimationView
            
        case .noPaymentMethod:
            currentView = noPaymentView
        }
        [noBubbleView,
         startingView,
         disconnectedView,
         connectingView,
         connectedView,
         noPaymentView,
            ].forEach { $0?.alpha = currentView == $0 ? 1 : 0 }
        [noBubbleAnimationView,
         startingAnimationView,
         connectingAnimationView,
         connectedAnimationView,
            ].filter { $0 != currentAnimationView }
            .forEach { $0?.stop() }
        if let currentAnimationView = currentAnimationView, !currentAnimationView.isAnimationPlaying {
            currentAnimationView.currentFrame = 0
            currentAnimationView.play()
        }
    }
    
    private func connectingAnimationFrame(for progress: Double) -> AnimationFrameTime {
        let firstProgressFrame: CGFloat = 22
        let lastProgressFrame: CGFloat = 49
        let lastAnimationFrame: CGFloat = 75
        guard progress > 0 else {
            return 0
        }
        guard progress < 1 else {
            return lastAnimationFrame
        }
        return firstProgressFrame + .init(.init(lastProgressFrame - firstProgressFrame) * progress)
    }
}
