// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

#if os(iOS)
import UIKit
#elseif os(macOS)
import Cocoa
#endif

class GradientView: UserInteractionDisablableView {
    struct Props {
        struct Point {
            let color: Color
            let location: Float
        }
        struct Vector {
            let startPoint: CGPoint
            let endPoint: CGPoint
        }
        enum Direction {
            case vertical
            case horizontal
            case custom(vector: Vector)
            
            var vector: Vector {
                switch self {
                case .vertical:
                    return .init(startPoint: .init(x: 0.5, y: 0),
                                 endPoint: .init(x: 0.5, y: 1))
                case .horizontal:
                    return .init(startPoint: .init(x: 0, y: 0.5),
                                 endPoint: .init(x: 1, y: 0.5))
                    
                case .custom(let vector):
                    return vector
                }
            }
        }
        
        let points: [Point]
        let direction: Direction
        
        static let zero: Self = .init(points: [], direction: .horizontal)
        
        static let connect: Self = .init(
            points: [ .init(color: .init(rgb: 0xFFFFFF), location: 0),
                      .init(color: .init(rgb: 0xEDEEF8), location: 1), ],
            direction: .vertical)
        
        static let mainButton: Self = .init(
            points: [ .init(color: .init(rgb: 0xEE2F8E), location: 0.13),
                      .init(color: .init(rgb: 0xB838C9), location: 1), ],
            direction: .horizontal)
        
        static let contactSupportButton: Self = .init(
            points: [ .init(color: .init(rgb: 0xC037C1), location: 0.13),
                      .init(color: .init(rgb: 0x64176F), location: 1), ],
            direction: .horizontal)
    }
    
    var props: Props = .zero { didSet { updateGradient() } }
    
    override init(frame: CGRect = .zero) {
        super.init(frame: frame)
        customInit()
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
        customInit()
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        gradientLayer.frame = bounds
    }
    
    private let gradientLayer = CAGradientLayer()
    private var needsUpdateGradient = false
    
    private func customInit() {
        backingLayer.insertSublayer(gradientLayer, at: 0)
        updateGradient()
    }
    
    private func setNeedsUpdateGradient() {
        guard !needsUpdateGradient else { return }
        needsUpdateGradient = true
        DispatchQueue.main.async {
            self.updateGradient()
        }
    }
    
    private func updateGradient() {
        gradientLayer.startPoint = props.direction.vector.startPoint
        gradientLayer.endPoint = props.direction.vector.endPoint
        gradientLayer.colors = props.points.map { $0.color.cgColor }
        gradientLayer.locations = props.points.map { .init(value: $0.location) }
        needsUpdateGradient = false
    }
}
