// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

#if os(iOS)
import UIKit
#elseif os(macOS)
import Cocoa
#endif

extension View {
    func animate(
        withDuration duration: TimeInterval,
        animations: @escaping (View) -> Void,
        completionHandler: (() -> Void)? = nil)
    {
        #if os(iOS)
        View.animate(withDuration: duration) {
            animations(self)
        }
        #elseif os(macOS)
        NSAnimationContext.runAnimationGroup({ context in
            context.duration = duration
            animations(animator())
        }, completionHandler: completionHandler)
        #endif
    }
}

class NetworkStateView: Label {
    enum State {
        case unavailable
        case available
    }
    
    var state: State = .available {
        didSet {
            guard state != oldValue else { return }
            stateUpdated(animated: true)
        }
    }
    
    override func awakeFromNib() {
        super.awakeFromNib()
        stateUpdated(animated: false)
        textColor = .white
        #if os(iOS)
        clipsToBounds = true
        #elseif os(macOS)
        #endif
        (self as View).backgroundColor = .clear
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        backingLayer.cornerRadius = bounds.height / 2
    }
    
    #if os(iOS)
    override func drawText(in rect: CGRect) {
        let insets: CGFloat = 10
        super.drawText(in: rect.inset(by: .init(top: 0, left: insets, bottom: 0, right: insets)))
    }
    #endif
    
    private func stateUpdated(animated: Bool) {
        switch state {
        case .unavailable:
            text = "Sorry, friends. Network not currently available."
            backingLayer.animate(.init(
                keyPath: \.backgroundColor,
                value: R.color.networkUnavailable()?.cgColor,
                duration: isHidden ? 0 : 0.3)
            )
            backingLayer.animate(.init(
                keyPath: \.opacity,
                value: 1,
                duration: 0.3)
            )
            isHidden = false
            
        case .available:
            text = "Network available. Try again!"
            
            guard animated else {
                alpha = 0
                isHidden = true
                return
            }
            
            backingLayer.animate(.init(
                keyPath: \.backgroundColor,
                value: R.color.networkAvailable()?.cgColor,
                duration: 0.3)
            )
            
            DispatchQueue.main.asyncAfter(deadline: .now() + 3) { [weak self] in
                guard let self = self, self.state == .available else { return }
                
                self.backingLayer.animate(.init(
                    keyPath: \.opacity,
                    value: 0,
                    duration: 0.3)
                ) { [weak self] in
                    guard let self = self, self.state == .available else { return }
                    self.isHidden = true
                }
            }
        }
    }
}
