// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

#if os(iOS)
import UIKit

typealias ViewController = UIViewController
typealias View = UIView
typealias Button = UIButton
typealias TextField = UITextField
typealias ImageView = UIImageView
typealias Label = UILabel
typealias UserInteractionDisablableView = View

typealias Color = UIColor
typealias Image = UIImage

extension UIButton {
    var title: String? {
        get { title(for: .normal) }
        set { setTitle(newValue, for: .normal) }
    }
}
#elseif os(macOS)
import Cocoa

typealias ViewController = NSWindowController
typealias View = NSView
typealias Button = NSButton
typealias TextField = NSTextField
typealias ImageView = NSImageView
typealias Label = NSTextField

typealias Color = NSColor
typealias Image = NSImage

class UserInteractionDisablableView: View {
    var isUserInteractionEnabled: Bool = true
    
    override func hitTest(_ point: NSPoint) -> NSView? {
        guard isUserInteractionEnabled, alpha > 0.01, !isHidden else { return nil }
        return super.hitTest(point)
    }
}

extension TextField {
    var text: String {
        get { stringValue }
        set { stringValue = newValue }
    }
}
#endif

extension ViewController {
    #if os(macOS)
    static let classInit: Void = {
        let selector: Selector
        selector = #selector(ViewController.windowDidLoad)
        guard
            let originalMethod = class_getInstanceMethod(ViewController.self, selector),
            let swizzledMethod = class_getInstanceMethod(ViewController.self, #selector(ViewController.viewDidLoad))
            else { return }
        method_exchangeImplementations(originalMethod, swizzledMethod)
    }()
    
    @objc func viewDidLoad() {
        viewDidLoad()
    }
    #endif
}

extension View {
    #if os(macOS)
    static let classInit: Void = {
        let selector: Selector
        selector = #selector(View.layout)
        guard
            let originalMethod = class_getInstanceMethod(View.self, selector),
            let swizzledMethod = class_getInstanceMethod(View.self, #selector(View.layoutSubviews))
            else { return }
        method_exchangeImplementations(originalMethod, swizzledMethod)
    }()
    
    @objc func layoutSubviews() {
        layoutSubviews()
    }
    
    var backgroundColor: Color? {
        get { Color(optCGColor: backingLayer.backgroundColor) }
        set { backingLayer.backgroundColor = newValue?.cgColor }
    }
    var alpha: CGFloat {
        get { alphaValue }
        set {
            alphaValue = newValue
        }
    }
    
    func bringSubviewToFront(_ view: View) {
        guard subviews.last != view else { return }
        var view = view
        sortSubviews({ view1, view2, rawPointer in
            switch rawPointer?.load(as: NSView.self) {
            case view1:
                return ComparisonResult.orderedDescending
            case view2:
                return ComparisonResult.orderedAscending
            default:
                return ComparisonResult.orderedSame
            }
        }, context: &view)
    }
    
    #endif
    
    var backingLayer: CALayer {
        // on iOS layer is non optional
        #if os(iOS)
        return layer
        #elseif os(macOS)
        if !wantsLayer {
            wantsLayer = true
        }
        return layer!
        #endif
    }
    
    func insertAtZeroPosition(_ view: View) {
        #if os(iOS)
        insertSubview(view, at: 0)
        #elseif os(macOS)
        addSubview(view, positioned: .below, relativeTo: nil)
        #endif
    }
}

extension Color {
    convenience init?(optCGColor: CGColor?) {
        guard let optCGColor = optCGColor else { return nil }
        self.init(cgColor: optCGColor)
    }
}
