// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

#if os(iOS)
import UIKit
#elseif os(macOS)
import Cocoa
#endif

class LoginTextField: TextField {
    private let errorBorderWidth: CGFloat = 1
    
    private let innerBorderView: View = {
        let view = UserInteractionDisablableView()
        view.backingLayer.borderColor = R.color.textFieldBorder()?.cgColor
        view.backingLayer.borderWidth = 1
        view.isUserInteractionEnabled = false
        return view
    }()
    
    var isError = false {
        didSet {
            updateLayerBorder()
        }
    }
    
    private var disabledOverlay: View = {
        let view = View()
        view.backgroundColor = R.color.disabledButtonOverlay()
        view.alpha = 0
        return view
    }()
    
    override var isEnabled: Bool {
        didSet {
            disabledOverlay.alpha = isEnabled ? 0 : 1
        }
    }
    
    #if os(iOS)    
    override func bringSubviewToFront(_ view: UIView) {
        super.bringSubviewToFront(view)
        super.bringSubviewToFront(disabledOverlay)
    }
    #elseif os(macOS)
    override func didAddSubview(_ subview: NSView) {
        super.didAddSubview(subview)
        bringSubviewToFront(disabledOverlay)
    }
    
    @discardableResult override func becomeFirstResponder() -> Bool {
        guard super.becomeFirstResponder() else { return false }
        if let textView = currentEditor() as? NSTextView, let textColor = textColor {
            textView.insertionPointColor = textColor
        }
        return true
    }
    #endif
    
    override func awakeFromNib() {
        super.awakeFromNib()
        backingLayer.cornerRadius = 3
        backingLayer.borderWidth = errorBorderWidth
        backingLayer.backgroundColor = R.color.textFieldBackground()?.cgColor
        updateLayerBorder()
        innerBorderView.backingLayer.cornerRadius = backingLayer.cornerRadius
        insertAtZeroPosition(innerBorderView)
        addSubview(disabledOverlay)
        #if os(iOS)
        attributedPlaceholder = .init(
            string: placeholder ?? "",
            attributes: [.foregroundColor: R.color.textFieldPlaceholder()!]
        )
        #elseif os(macOS)
        if let cell = (cell as? NSTextFieldCell) {
            let paragraphStyle = NSMutableParagraphStyle()
            paragraphStyle.alignment = .center
            cell.placeholderAttributedString = .init(
                string: cell.placeholderString ?? "",
                attributes: [.foregroundColor: R.color.textFieldPlaceholder()!,
                             .font: font!,
                             .paragraphStyle: paragraphStyle]
            )
        }
        focusRingType = .none
        #endif
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        innerBorderView.frame = bounds.insetBy(dx: errorBorderWidth, dy: errorBorderWidth)
        disabledOverlay.frame = bounds
    }
    
    private func updateLayerBorder() {
        backingLayer.borderColor = (isError ? R.color.error() : R.color.loginBackground())?.cgColor
    }
}
