// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import UIKit
import AVKit
import Connectivity

class ConnectionViewController: UIViewController, iPadStretchable {
    @IBOutlet weak var backgroundView: GradientView!
    @IBOutlet weak var installCertificateView: UIView!
    @IBOutlet weak var connectionView: ConnectionView!
    @IBOutlet weak var networkStateView: NetworkStateView!
    @IBOutlet var stretchOniPadConstraints: [NSLayoutConstraint]!
    @IBOutlet var navigationButtonsToConnectButtonWidthConstraint: NSLayoutConstraint!
    
    var observers = [AnyObject]()
    
    var connectionManager: ConnectionManager!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        adaptForiPad { multiplier in
            navigationButtonsToConnectButtonWidthConstraint.constant = 50 * multiplier
        }
        backgroundView.props = .connect
        updateSSLValid(true, animated: false)
        connectionManager.invalidSSLHandler = { [weak self] in
            self?.updateSSLValid(false, animated: true)
        }

        observers = [
            connectionManager.addStatusChangedHandler(statusHandler: .init({ [weak self] status in
                self?.statusChanged(status)
            })),
            connectionManager.observe(\.error) { [weak self] manager, _ in
                self?.handleError(manager.error)
            },
            NotificationCenter.default.addObserver(
                forName: UIApplication.didBecomeActiveNotification,
                object: nil,
                queue: .main)
            { [weak self] _ in self?.applicationDidBecomeActive() },
        ]
        connectionManager.initialize()
        connectivity.whenConnected = { [weak self] _ in
            self?.networkStateView.state = .available
        }
        connectivity.framework = .network
        connectivity.startNotifier()
    }
    
    override var prefersStatusBarHidden: Bool { false }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        appearing = true
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        appearing = false
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        // containerView
        switch segue.destination {
        case let destination as AVPlayerViewController:
            playerController = destination
            
        default: break
        }
        // performSegues
        if let typedInfo = R.segue.connectionViewController.error(segue: segue) {
            typedInfo.destination.error = sender as? Error
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
    }
    
    @IBAction func connectButtonTap() {
        connectionManager.toggleStatus()
    }
    
    @IBAction func launchBubbleButtonTap() {
        connectionManager.createAccountPlanAndLaunch()
    }
    
    @IBAction func logOutButtonTap() {
        connectionManager.logOut()
        navigationController?.popToRootViewController(animated: true)
    }
    
    @IBAction func myBubbleButtonTap() {
        UIApplication.shared.open(connectionManager.myBubbleURL!)
    }
    
    @IBAction func supportButtonTap() {
        UIApplication.shared.open(URL(string: "https://support.getbubblenow.com/")!)
        updateSSLValid(true, animated: true)
    }
    
    @IBAction func getCertButtonTap() {
        guard let caCertURL = connectionManager.caCertURL else { return }
        #if !DEBUG
        UIApplication.shared.open(caCertURL)
        #endif
        updateSSLValid(true, animated: true)
    }
    
    @IBAction func addCardButtonTap() {
        UIApplication.shared.open(connectionManager.paymentPageURL!)        
    }
    
    private let guideURLString = "https://github.com/getbubblenow/bubble-docs/blob/master/cert_instructions/ios_screenshots/iOS-CA-Certificate-Installation.mp4?raw=true"
    
    private var appearing = false
    private let connectivity = Connectivity(shouldUseHTTPS: false)
    
    private weak var playerController: AVPlayerViewController! {
        didSet {
            playerController?.player = .init(url: URL(string: guideURLString)!)
        }
    }
    
    private func updateSSLValid(_ valid: Bool, animated: Bool) {
        UIView.animate(withDuration: animated ? 0.3 : 0) {
            self.installCertificateView.alpha = valid ? 0 : 1
            self.connectionView.alpha = valid ? 1 : 0
        }
    }
    
    private func statusChanged(_ status: ConnectionManager.Status) {
        UIView.animate(withDuration: appearing ? 0 : 0.3,
                       delay: 0,
                       options: [.beginFromCurrentState],
                       animations: { self.connectionView.props = .init(status: status) },
                       completion: nil)
    }
    
    private func handleError(_ error: Error?) {
        guard let error = error else { return }
        switch error as? NetworkManagerError {
        case .tokenExpired:
            logOutButtonTap()
            return
            
        case .networkError(let code) where code == NSURLErrorCancelled:
            if !connectivity.isConnected {
                networkStateView.state = .unavailable
                return
            }
            
        default: break
        }
        #if DEBUG
        wg_log(error: error)
        #else
        performSegue(withIdentifier: R.segue.connectionViewController.error, sender: error)
        #endif
    }
    
    private func applicationDidBecomeActive() {
        guard connectionManager.status == .noPaymentMethod else { return }
        connectionManager.checkPaymentMethod()
    }
}
