// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import UIKit
import MessageUI

class ErrorViewController: UIViewController {
    @IBOutlet weak var button: GradientButton!
    @IBOutlet weak var titleLabel: UILabel!
    
    var error: Error?

    override func viewDidLoad() {
        super.viewDidLoad()
        button.gradientProps = .contactSupportButton
        wg_log(error: currentError, message: "\(Self.self) \(#function)")
        
        #if DEBUG
        createLogsFile { [weak self] logsURL in
            guard let self = self else { return }
            self.shareWithActivity(logsURL: logsURL)
        }
        #endif
    }
    
    @IBAction func buttonTap() {
        createLogsFile { [weak self] logsURL in
            guard let self = self else { return }
            do {
                try self.composeMail(logsURL: logsURL)
            } catch {
                self.shareWithActivity(logsURL: logsURL)
            }
        }
    }
    
    private var currentError: Error {
        error ?? NSError(description: "ErrorViewController empty error")
    }
    
    private func composeMail(logsURL: URL?) throws {
        guard MFMailComposeViewController.canSendMail() else { throw NSError(description: "Mail app not available") }
        let controller = MFMailComposeViewController()
        controller.popoverPresentationController?.sourceView = titleLabel
        controller.mailComposeDelegate = self
        var messageBody = "Please describe what happened or just send us the attached logs"
        if let logsURL = logsURL, let logData = try? Data(contentsOf: logsURL) {
            controller.addAttachmentData(logData, mimeType: "text/plain", fileName: logsURL.lastPathComponent)
        } else {
            messageBody += "\n\(currentError.fixedLocalizedDescription)"
        }
        controller.setMessageBody(messageBody, isHTML: false)
        controller.setSubject("iOS app Bubble error")
        controller.setToRecipients([Constants.supportEmail])
        present(controller, animated: true)
    }
    
    private func shareWithActivity(logsURL: URL?) {
        let item: Any = logsURL ?? currentError.fixedLocalizedDescription
        let controller = UIActivityViewController(activityItems: [item], applicationActivities: nil)
        controller.popoverPresentationController?.sourceView = titleLabel
        let originalTitleText = titleLabel.text
        titleLabel.text = "Please send logs file to\n\(Constants.supportEmail)\nAddress will be copied to your clipboard if you save the file"
        controller.completionWithItemsHandler = { [weak self] _, completed, _, _ in
            guard let self = self else { return }
            if completed {
                UIPasteboard.general.string = Constants.supportEmail
                self.dismiss(animated: true)
            } else {
                self.titleLabel.text = originalTitleText
            }
        }
        present(controller, animated: true)
    }
    
    private func createLogsFile(completion: @escaping (URL?) -> Void) {
        DispatchQueue(label: "createLogs").async {
            let dateFormatter = ISO8601DateFormatter()
            dateFormatter.formatOptions = [.withFullDate, .withTime, .withTimeZone]
            let timeStampString = dateFormatter.string(from: Date())
            let destinationURL = URL(fileURLWithPath: NSTemporaryDirectory())
                .appendingPathComponent("bubble-log-\(timeStampString).txt")
            try? FileManager.default.removeItem(at: destinationURL)
            var result: URL?
            if Logger.global?.writeLog(to: destinationURL.path) == true {
                result = destinationURL
            }
            DispatchQueue.main.async {
                completion(result)
            }
        }
    }
}

extension ErrorViewController: MFMailComposeViewControllerDelegate {
    func mailComposeController(_ controller: MFMailComposeViewController, didFinishWith result: MFMailComposeResult, error: Error?) {
        controller.dismiss(animated: true) { [weak self] in
            switch result {
            case .cancelled, .failed: break
                
            case .sent, .saved:
                self?.dismiss(animated: true)
                
            @unknown default: break
            }
        }
    }
}
