// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 bubble. All Rights Reserved.

import UIKit
import SwiftyUserDefaults

class LoginViewController: UIViewController, iPadStretchable {
    enum State {
        case idle
        case waitingForResponse
    }
    
    @IBOutlet weak var userNameTextField: LoginTextField!
    @IBOutlet weak var passwordTextField: LoginTextField!
    @IBOutlet weak var errorLabel: ErrorLabel!
    @IBOutlet weak var signInButton: UIButton!
    @IBOutlet weak var customHostnameButton: UIButton!
    @IBOutlet weak var keyboardPositionConstraint: NSLayoutConstraint!
    @IBOutlet var controls: [UIControl]!
    @IBOutlet var stretchOniPadConstraints: [NSLayoutConstraint]!
    
    override var prefersStatusBarHidden: Bool { true }
    
    private var defaultsDisposable: DefaultsDisposable?
    private var customURL: URL? {
        didSet {
            customHostnameButton.isHidden = customURL == nil
            guard customURL != oldValue else { return }
            error = nil
            guard let customURL = customURL else { return }
            customHostnameButton.setTitle("Custom hostname: \(customURL)", for: .normal)
        }
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        adaptForiPad()
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        keyboardNotifier.enabled = true
        if let networkManager = BubbleKeychain.shared.currentSageManager {
            performSegue(withIdentifier: R.segue.loginViewController.pushWithoutAnimation, sender: networkManager)
        } else {
            credentials = .loadSaved()
        }
        
        defaultsDisposable = Defaults
            .observe(\.settingsBundleHostname) { [weak self] change in
                self?.checkSettingsHostname(change.newValue ?? "")
            }
        checkSettingsHostname()
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        guard
            let typedInfo = R.segue.loginViewController.push(segue: segue) ??
                R.segue.loginViewController.pushWithoutAnimation(segue: segue),
            let sageManager = sender as? SageManager
            else { return }
        typedInfo.destination.connectionManager = .init(sageManager: sageManager)
    }
    
    override func viewDidDisappear(_ animated: Bool) {
        keyboardNotifier.enabled = false
        defaultsDisposable?.dispose()
        super.viewDidDisappear(animated)
    }
    
    @IBAction func signInButtonTap() {
        guard let credentials = credentials else { return }
        state = .waitingForResponse
        sageFabric.login(credentials: credentials, customURL: customURL) { result in
            DispatchQueue.main.async {
                self.state = .idle
                switch result {
                case .success(let sageManager):
                    BubbleKeychain.shared.currentSageManager = sageManager
                    self.performSegue(withIdentifier: R.segue.loginViewController.push, sender: sageManager)
                    
                case .failure(let error):
                    self.error = error
                }
            }
        }
    }
    
    @IBAction func signUpButtonTap() {
        state = .waitingForResponse
        sageFabric.getRegisterURL { result in
            DispatchQueue.main.async {
                self.state = .idle
                switch result {
                case .success(let url):
                    UIApplication.shared.open(url)
                    
                case .failure(let error):
                    self.error = error
                }
            }
        }
    }
    
    @IBAction func forgotPasswordButtonTap() {
        state = .waitingForResponse
        sageFabric.getForgotPasswordURL { result in
            DispatchQueue.main.async {
                self.state = .idle
                switch result {
                case .success(let url):
                    UIApplication.shared.open(url)
                    
                case .failure(let error):
                    self.error = error
                }
            }
        }
    }
    
    @IBAction func customHostnameButtonTap() {
        presentUpdateHostnameAlert()
    }
    
    private let sageFabric = SageFabric()
    private var credentials: Credentials? {
        get {
            Credentials(username: userNameTextField.text,
                        password: passwordTextField.text)
        }
        set {
            userNameTextField.text = newValue?.username
            passwordTextField.text = newValue?.password
            signInButton.isEnabled = newValue != nil
        }
    }
    private var state: State = .idle {
        didSet {
            stateUpdated(state)
        }
    }
    private var error: Error? {
        didSet {
            errorLabel.error = error
            
            [userNameTextField,
             passwordTextField,
                ].forEach {
                    $0?.isError = error != nil
            }
        }
    }
    private lazy var keyboardNotifier = KeyboardNotifier(parentView: view, constraint: keyboardPositionConstraint, isAnchor: true)
    
    private func stateUpdated(_ state: State) {
        let controlsEnabled: Bool
        switch state {
        case .idle:
            controlsEnabled = true
            
        case .waitingForResponse:
            controlsEnabled = false
        }
        controls.forEach {
            $0.isEnabled = controlsEnabled
            $0.resignFirstResponder()
        }
        error = nil
    }
    
    private func checkSettingsHostname(_ newValue: String? = nil) {
        let settingsBundleHostname = newValue ?? Defaults.settingsBundleHostname
        var customURL: URL?
        defer {
           self.customURL = customURL
        }
        guard !settingsBundleHostname.isEmpty else { return }
        if let url = URL(sslSafe: settingsBundleHostname) {
            customURL = url
            return
        }
        presentUpdateHostnameAlert()
    }
    
    private func presentUpdateHostnameAlert() {
        class Handler: NSObject {
            let action: UIAlertAction
            var text: String
            var url: URL? { URL(sslSafe: text) }
            
            init(_ action: UIAlertAction, text: String) {
                self.action = action
                self.text = text
                super.init()
                updateAction()
            }
            
            @objc func alertTextFieldEditingChanged(textField: UITextField) {
                text = textField.text ?? ""
                updateAction()
            }
            
            private func updateAction() {
                action.isEnabled = url != nil
            }
        }
        let messageDetails: String
        if let customURL = customURL {
            messageDetails = "\(customURL) is used as a hostname."
        } else {
            messageDetails = "\(Defaults.settingsBundleHostname) entered in settings is invalid url."
        }
        let alert = UIAlertController(
            title: "Custom bubble hostname",
            message: messageDetails + " Update it or revert back to default settings",
            preferredStyle: .alert)
        var handler: Handler!
        handler = Handler(.init(title: "Save", style: .default, handler: { [weak self] _ in
            Defaults.settingsBundleHostname = handler.text
            self?.customURL = handler.url
        }), text: Defaults.settingsBundleHostname)
        alert.addTextField { textField in
            textField.text = Defaults.settingsBundleHostname
            textField.addTarget(
                handler,
                action: #selector(Handler.alertTextFieldEditingChanged(textField:)),
                for: .editingChanged
            )
        }
        alert.addAction(.init(title: "Revert", style: .destructive, handler: { [weak self] _ in
            Defaults.settingsBundleHostname = ""
            self?.customURL = nil
        }))
        alert.addAction(handler.action)
        present(alert, animated: true)
    }
}

extension LoginViewController: UITextFieldDelegate {
    func textFieldDidBeginEditing(_ textField: UITextField) {
        error = nil
    }
    
    @IBAction func textFieldEditingChanged() {
        signInButton.isEnabled = credentials != nil
    }
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        guard let currentIndex = controls.firstIndex(of: textField),
            case let nextIndex = currentIndex + 1,
            nextIndex < controls.count
            else { return false }
        switch controls[nextIndex] {
        case let nextControl as UITextField:
            nextControl.becomeFirstResponder()
            
        case let nextControl as UIButton:
            nextControl.sendActions(for: .touchUpInside)
            
        default:
            return false
        }
        return true
    }
}
