// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Cocoa
import ServiceManagement

class AppDelegate: NSObject, NSApplicationDelegate {
    
    var statusItemController: StatusItemController?
    
    var onAppDeactivation: (() -> Void)?
    
    var appPrivacyWindowController: AppPrivacyWindowController?
    var currentWindowController: NSWindowController?
    var connectionWindowController: ConnectionWindowController?
    var preferencesWindowController: PreferencesWindowController?
    
    #if DEVELOPER_ID
    let flexrouterInstaller = FlexrouterInstaller()
    #endif
    
    func applicationWillFinishLaunching(_ notification: Notification) {
        UserDefaults.standard.set(true, forKey: "NSDisabledDictationMenuItem")
        UserDefaults.standard.set(true, forKey: "NSDisabledCharacterPaletteMenuItem")
        NSApp.setActivationPolicy(.prohibited)
        View.classInit
    }
    
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        Logger.configureGlobal(tagged: "APP", withFilePath: FileManager.logFileURL?.path)
           
        var isLaunchedAtLogin = false
        if let appleEvent = NSAppleEventManager.shared().currentAppleEvent {
            isLaunchedAtLogin = LaunchedAtLoginDetector.isLaunchedAtLogin(openAppleEvent: appleEvent)
        }
        
        setDockIconAndMainMenuVisibility(isVisible: !isLaunchedAtLogin)
        updateCurrentWindow()
        #if DEVELOPER_ID
        DispatchQueue.main.asyncAfter(deadline: .now() + 1) {
            self.flexrouterInstaller.installIfNeeded()            
        }
        #endif
    }
    
    func updateCurrentWindow(sageManager: SageManager? = nil) {
        currentWindowController?.close()
        var shouldLaunchAtLogin = false
        if let sageManager = sageManager ?? BubbleKeychain.shared.currentSageManager {
            shouldLaunchAtLogin = true
            (connectionWindowController, statusItemController)
                = initializeConnectionAndStatusItemControllers(sageManager: sageManager)
            currentWindowController = connectionWindowController
        } else {
            currentWindowController = initializeLoginController()
            connectionWindowController = nil
            statusItemController = nil
        }
        registerLoginItem(shouldLaunchAtLogin: shouldLaunchAtLogin)
        currentWindowController?.window?.makeKeyAndOrderFront(self)
    }
    
    func applicationShouldHandleReopen(_ sender: NSApplication, hasVisibleWindows: Bool) -> Bool {
        if let appleEvent = NSAppleEventManager.shared().currentAppleEvent {
            if LaunchedAtLoginDetector.isReopenedByLoginItemHelper(reopenAppleEvent: appleEvent) {
                return false
            }
        }
        if hasVisibleWindows {
            return true
        }
        updateCurrentWindow()
        return false
    }
    
    @objc func logOut() {
        connectionWindowController?.logOut()
    }
    
    @IBAction func quit(_ sender: Any? = nil) {
        if let manageWindow = currentWindowController?.window, manageWindow.attachedSheet != nil {
            NSApp.activate(ignoringOtherApps: true)
            manageWindow.orderFront(self)
            return
        }
        registerLoginItem(shouldLaunchAtLogin: false)
        switch connectionWindowController?.connectionManager.status {
        case .active, .activating:
            let alert = NSAlert()
            alert.messageText = "Bubble is exiting with an active tunnel"
            alert.informativeText = "The tunnel will remain active after exiting. You may disable it by reopening this application or through the Network panel in System Preferences."
            NSApp.activate(ignoringOtherApps: true)
            if let manageWindow = currentWindowController?.window {
                manageWindow.orderFront(self)
                alert.beginSheetModal(for: manageWindow) { _ in
                    NSApp.terminate(nil)
                }
            } else {
                alert.runModal()
                NSApp.terminate(nil)
            }
            
        default:
            NSApp.terminate(nil)
        }
    }
    
    func applicationShouldTerminate(_ sender: NSApplication) -> NSApplication.TerminateReply {
        guard
            let connectionManager = connectionWindowController?.connectionManager,
            connectionManager.status == .active || connectionManager.status == .activating
            else { return .terminateNow }
        guard let appleEvent = NSAppleEventManager.shared().currentAppleEvent else {
            return .terminateNow
        }
        guard MacAppStoreUpdateDetector.isUpdatingFromMacAppStore(quitAppleEvent: appleEvent) else {
            return .terminateNow
        }
        let alert = NSAlert()
        alert.messageText = "App Store would like to update Bubble"
        alert.informativeText = "Please disconnect from the bubble and then continue updating in App Store."
        NSApp.activate(ignoringOtherApps: true)
        if let manageWindow = currentWindowController?.window {
            alert.beginSheetModal(for: manageWindow) { _ in }
        } else {
            alert.runModal()
        }
        return .terminateCancel
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ application: NSApplication) -> Bool {
        DispatchQueue.main.asyncAfter(deadline: .now() + .milliseconds(200)) { [weak self] in
            self?.setDockIconAndMainMenuVisibility(isVisible: false)
        }
        return false
    }
    
    private func setDockIconAndMainMenuVisibility(isVisible: Bool, completion: (() -> Void)? = nil) {
        let currentActivationPolicy = NSApp.activationPolicy()
        let newActivationPolicy: NSApplication.ActivationPolicy = isVisible ? .regular : .accessory
        guard currentActivationPolicy != newActivationPolicy else {
            if newActivationPolicy == .regular {
                NSApp.activate(ignoringOtherApps: true)
            }
            completion?()
            return
        }
        if newActivationPolicy == .regular && NSApp.isActive {
            onAppDeactivation = {
                NSApp.setActivationPolicy(.regular)
                NSApp.activate(ignoringOtherApps: true)
                completion?()
            }
            NSApp.setActivationPolicy(.prohibited)
        } else {
            NSApp.setActivationPolicy(newActivationPolicy)
            if newActivationPolicy == .regular {
                NSApp.activate(ignoringOtherApps: true)
            }
            completion?()
        }
    }
    
    func applicationDidResignActive(_ notification: Notification) {
        onAppDeactivation?()
        onAppDeactivation = nil
    }
    
    private func initializeLoginController() -> LoginWindowController {
        let loginWindowController = LoginWindowController.controllerFromNib()
        loginWindowController.onSignIn = { [unowned self] sageManager in
            BubbleKeychain.shared.currentSageManager = sageManager
            self.updateCurrentWindow(sageManager: sageManager)
        }
        loginWindowController.onPrefsRequest = { [unowned self] in
            self.showPreferences()
        }
        return loginWindowController
    }
    
    private func initializeConnectionAndStatusItemControllers(sageManager: SageManager) -> (ConnectionWindowController, StatusItemController) {
        let connectionWindowController = ConnectionWindowController.controllerFromNib()
        connectionWindowController.connectionManager = .init(sageManager: sageManager)
        connectionWindowController.onLogOut = { [unowned self] in
            BubbleKeychain.shared.currentSageManager = nil
            self.updateCurrentWindow()
        }
        let statusItemController = StatusItemController(connectionManager: connectionWindowController.connectionManager)
        statusItemController.statusItem.menu = StatusMenu(connectionManager: connectionWindowController.connectionManager)
        return (connectionWindowController, statusItemController)
    }

    private var errorController: ErrorWindowController!
    
    @objc func showPreferences() {
        if preferencesWindowController == nil {
            preferencesWindowController = PreferencesWindowController.controllerFromNib()
        }
        preferencesWindowController?.window?.makeKeyAndOrderFront(self)
    }

    func presentErrorController(_ error: Error, forcePresent: Bool = false) {
        var presentNeeded = forcePresent
        #if !DEBUG
        presentNeeded = true
        #endif
        if presentNeeded {
            errorController = .controllerFromNib()
            errorController.error = error
            errorController.window?.makeKeyAndOrderFront(self)
        } else {
            wg_log(error: error)
        }
    }
}

extension AppDelegate {
    @objc func aboutClicked() {
        var appVersion = Bundle.main.infoDictionary?["CFBundleShortVersionString"] as? String ?? "Unknown"
        if let appBuild = Bundle.main.infoDictionary?["CFBundleVersion"] as? String {
            appVersion += " (\(appBuild))"
        }
        let appVersionString = [tr(format: "macAppVersion (%@)", appVersion),
                                tr(format: "macGoBackendVersion (%@)", WIREGUARD_GO_VERSION)
            ].joined(separator: "\n")
        NSApp.activate(ignoringOtherApps: true)
        NSApp.orderFrontStandardAboutPanel(options: [
            .applicationVersion: appVersionString,
            .version: "",
            .credits: ""
        ])
    }
}

@discardableResult
func registerLoginItem(shouldLaunchAtLogin: Bool) -> Bool {
    let appId = Bundle.main.bundleIdentifier!
    let helperBundleId = "\(appId).login-item-helper"
    return SMLoginItemSetEnabled(helperBundleId as CFString, shouldLaunchAtLogin)
}

extension NSWindow.FrameAutosaveName {
    static let shared: NSWindow.FrameAutosaveName = .init("Position")
}
