// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation
import SecurityFoundation
import ServiceManagement

class FlexrouterInstaller {
    private let client = XPCClient()
    
    func installIfNeeded() {
        getInstalledFlexrouterVersion { [weak self] installedVersion in
            guard let installedVersion = installedVersion else {
                self?.installFlexrouter()
                return
            }
            self?.getLatestFlexrouterVersion { latestVersion in
                if latestVersion != installedVersion {
                    self?.installFlexrouter()
                } else {
                    wg_log(.info, message: "flexrouter is up-to-date")
                }
            }
        }
    }
    
    private func installFlexrouter() {
        do {
            try initializeSMJob()
            client.install {
                wg_log(.info, message: "flexrouter is updated")
            } unexpectedHandler: { output in
                wg_log(
                    .info,
                    message: "install flexrouter unexpected output: \n\(output)"
                )
            }
        } catch {
            wg_log(error: error, message: "flexrouter initialization error")
        }
    }
    
    private func initializeSMJob() throws {
        var auth: AuthorizationRef?
        try tryOSStatus(
            AuthorizationCreate(nil, nil, [], &auth),
            description: "Create authorization"
        )
        var error: Unmanaged<CFError>?
        guard
            SMJobBless(
                kSMDomainSystemLaunchd,
                FlexrouterConstants.helperMachLabel as CFString,
                auth,
                &error
            ) == false
        else { return }
        throw error?.takeRetainedValue()
        ?? NSError(description: "SMJobBless unknown error")
    }
    
    private func getLatestFlexrouterVersion(completion: @escaping (String) -> Void) {
        let session = URLSession(configuration: .default)
        session.dataTask(
            with: FlexrouterConstants.latestVersionURL
        ) { [weak self] data, _, _ in
            data.map {
                guard let latestVersion = self?.parseVersionFromResponse($0)
                else { return }
                completion(latestVersion)
            }
        }.resume()
    }
    
    private func getInstalledFlexrouterVersion(completion: @escaping (String?) -> Void) {
        DispatchQueue(label: "process").async { completion({ [weak self] in
            let installationPath = FlexrouterConstants.installationURL.path
            guard
                FileManager.default.fileExists(
                    atPath: installationPath
                )
            else { return nil }
            let process = Process()
            process.launchPath = installationPath
            process.arguments = ["--version"]
            let standardOutput = Pipe()
            process.standardOutput = standardOutput
            
            process.launch()
            process.waitUntilExit()
            return self?.parseVersionFromResponse(
                standardOutput.fileHandleForReading.availableData
            )
        }())
        }
    }
    
    private func parseVersionFromResponse(_ data: Data) -> String? {
        guard
            let output = String(
                data: data,
                encoding: .utf8
            ),
            !output.isEmpty,
            let regex = try? NSRegularExpression(
                pattern: "(.+\\s)?((\\d\\.)+\\d+)"
            ),
            case let versionRangeIndex = 2,
            let match = regex.firstMatch(
                in: output,
                options: [],
                range: NSRange(
                    location: 0,
                    length: output.count
                )
            ),
            versionRangeIndex < match.numberOfRanges
        else { return nil }
        let version = (output as NSString)
            .substring(with: match.range(at: versionRangeIndex))
        return version
    }
}
