// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation

class InstallerImplementation: NSObject, Installer {
    
    var client: InstallationClient?
    
    func install(flexrouterPassword: String) {
        uninstall()
        downloadScript { [weak self] in
            self?.runInstallScript(script: $0, flexrouterPassword: flexrouterPassword)
        }
    }
    
    func uninstall() {
        _ = run(script: FlexrouterConstants.uninstallScriptURL)
    }
    
    private func runInstallScript(script: URL, flexrouterPassword: String) {
        let process = Process()
        process.environment = [
            "PATH": "/usr/bin:/bin:/usr/sbin",
            "BUBBLE_FR_PASS": flexrouterPassword,
        ]
        process.launchPath = "/bin/bash"
        process.arguments = ["-c", "sh " + script.path]
        
        let standardOutput = Pipe()
        process.standardOutput = standardOutput
        
        let standardError = Pipe()
        process.standardError = standardError
        
        process.terminationHandler = { [self] _ in
            guard
                let standardOutput = standardOutput.readAvailableDataString,
                let standardError = standardError.readAvailableDataString
            else { return }
            let output = """
            --------------------------
            standardOutput
            --------------------------
            \(standardOutput)
            --------------------------
            standardError
            --------------------------
            \(standardError)
            --------------------------
            """
            if standardOutput.contains("bubble-flexrouter successfully installed"),
               standardError.contains("Initialization completed successfully")
            {
                client?.finishedSuccessfully()
            } else {
                client?.finishedWithUnexpectedOutput(output)
            }
        }
        process.launch()
        process.waitUntilExit()
    }
    
    private func downloadScript(completion: @escaping (URL) -> Void) {
        let config = URLSessionConfiguration.default
        config.requestCachePolicy = .reloadIgnoringLocalCacheData
        config.urlCache = nil
        let session = URLSession(configuration: config)
        session.dataTask(with: FlexrouterConstants.installScriptURL) { data, _, _ in
            let scriptURL = URL(fileURLWithPath: NSTemporaryDirectory()).appendingPathComponent("script.sh")
            (try? data?.write(to: scriptURL))?.map {
                completion(scriptURL)
            }
        }.resume()
    }
    
    private func run(script: URL) -> String? {
        guard FileManager.default.fileExists(atPath: script.path) else { return nil }
        return run("/bin/bash", with: ["-c", "sh \(script.path)"])
    }
    
    private func run(_ command: String, with arguments: [String] = []) -> String {
        let process = Process()
        process.launchPath = command
        process.arguments = arguments
        let outputPipe = Pipe()
        process.standardOutput = outputPipe
        process.launch()
        process.waitUntilExit()
        let outputData = outputPipe.fileHandleForReading.readDataToEndOfFile()
        let output = String(decoding: outputData, as: UTF8.self)
        return output
    }
}

extension Pipe {
    var readAvailableDataString: String? {
        String(
            data: fileHandleForReading.availableData,
            encoding: .utf8
        )
    }
}
