// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Cocoa
import Carbon

class AddCardWindowController: NSWindowController {
    enum State {
        case idle
        case waitingForResponse
    }
    
    @IBOutlet weak var cancelButton: NSButton!
    @IBOutlet weak var addCardButton: NSButton!
    @IBOutlet weak var infoTextField: NSTextField!
    @IBOutlet weak var cardTextField: LoginTextField!
    @IBOutlet weak var dateTextField: LoginTextField!
    @IBOutlet weak var cvcTextField: LoginTextField!
    @IBOutlet weak var zipCodeTextField: LoginTextField!
    
    var connectionManager: ConnectionManager!
    
    private lazy var textFields: [LoginTextField] = [
        cardTextField,
        dateTextField,
        cvcTextField,
        zipCodeTextField,
    ]
    
    private var content = PaymentValidator.Content(card: .empty, cardBrand: nil)
    private var error: Error? {
        didSet {
            errorUpdated(error)
        }
    }
    private var state: State = .idle {
        didSet {
            stateUpdated(state)
        }
    }
    
    override func windowDidLoad() {
        super.windowDidLoad()
        error = nil
        styleCancelButton()
        textFields.forEach {
            $0.delegate = self
        }
        addCardButton.isEnabled = false
    }
    
    @IBAction func cancelButtonTap(_ sender: Any) {
        window?.sheetParent?.endSheet(window!, returnCode: .cancel)
    }
    
    @IBAction func addCardButtonTap(_ sender: Any) {
        state = .waitingForResponse
		//swiftlint:disable:next trailing_closure
        connectionManager.addCard(content.card, completionHandler: { [weak self] result in
            guard let self = self, let window = self.window else { return }
            self.state = .idle
            switch result {
            case .failure(let error):
                switch error {
                case .unexpected(let error) where error is ConnectionManager.StripeResponse.Error:
                    self.error = error
                    
                case .networkError(let error):
                    switch error {
                    case .serverErrors(let errors) where !errors.isEmpty:
                        self.error = error
                        return
                        
                    default: break
                    }
                    fallthrough
                    
                default:
                    (NSApp.delegate as? AppDelegate)?.presentErrorController(error)
                }
                
            case .success:
                window.sheetParent?.endSheet(window, returnCode: .OK)
            }
        })
    }
    
    private func styleCancelButton() {
        cancelButton.backingLayer.masksToBounds = false
        cancelButton.backingLayer.shadowColor = Color.black.withAlphaComponent(0.25).cgColor
        cancelButton.backingLayer.shadowRadius = 5
        cancelButton.backingLayer.shadowOpacity = 1
        cancelButton.backingLayer.shadowOffset = .zero
        cancelButton.backingLayer.backgroundColor = .white
        cancelButton.backingLayer.cornerRadius = cancelButton.bounds.height / 2
    }
    
    private func stateUpdated(_ state: State) {
        let controlsEnabled: Bool
        switch state {
        case .idle:
            controlsEnabled = true
            
        case .waitingForResponse:
            controlsEnabled = false
        }
        (textFields + [addCardButton, cancelButton])
            .forEach { $0.isEnabled = controlsEnabled }
        window?.makeFirstResponder(nil)
    }
    
    private func errorUpdated(_ error: Error?) {
        if let error = error {
            infoTextField.text = error.localizedDescription
            infoTextField.textColor = R.color.error()
        } else {
            infoTextField.text = """
                There is no charge for the first
                month of your Bubble.
                """
            infoTextField.textColor = R.color.gryText()
        }
    }
}

extension AddCardWindowController: NSTextFieldDelegate {
    func controlTextDidChange(_ obj: Notification) {
        guard let textField = obj.object as? LoginTextField else { return }
        textDidChange(textField)
    }
    
    func textDidChange(_ textField: LoginTextField) {
        error = nil
        let kind: PaymentValidator.Kind
        switch textField {
        case cardTextField:
            kind = .cardNumber
            
        case dateTextField:
            kind = .date
            
        case cvcTextField:
            kind = .cvc
            
        case zipCodeTextField:
            kind = .zipCode
            
        default: return
        }
        let result = PaymentValidator.validate(
            string: textField.text,
            kind: kind,
            content: &content
        )
        if result.failed {
            NSSound.beep()
        }
        if result.state == .complete, !result.failed {
            self.textField(textField, doCommandBy: #selector(NSResponder.insertTab))
        }
        textField.isError = result.state == .invalid
        textField.text = result.outputString
        addCardButton.isEnabled = content.valid
    }
    
    func control(
        _ control: NSControl,
        textView: NSTextView,
        doCommandBy commandSelector: Selector)
        -> Bool
    {
        guard let textField = control as? LoginTextField else { return false }
        return self.textField(textField, doCommandBy: commandSelector)
    }
    
    @discardableResult private func textField(
        _ textField: LoginTextField,
        doCommandBy commandSelector: Selector)
        -> Bool
    {
        guard let currentIndex = textFields.firstIndex(of: textField) else { return false }
        func safeIndex(_ index: Int) -> Int {
            (index + textFields.count) % textFields.count
        }
        switch commandSelector {
        case #selector(NSResponder.insertNewline(_:)):
            guard addCardButton.isEnabled else { fallthrough }
            addCardButton.performClick(control)
            return true
            
        case #selector(NSResponder.insertTab):
            if textField == dateTextField, textField.text.count == 1 {
                dateTextField.text = "0\(dateTextField.text)/"
                return true
            }
            var nextIndex = currentIndex + 1
            if nextIndex >= textFields.count, !addCardButton.isEnabled {
                nextIndex = 0
            }
            if nextIndex < textFields.count {
                textFields[nextIndex].becomeFirstResponder(selectionPosition: .end)
            } else {
                addCardButton.performClick(textField)
            }
            return true
            
        case #selector(NSResponder.deleteBackward):
            guard textField.selectionPosition != .start else {
                guard currentIndex > 0 else { break }
                let prevTextField = textFields[currentIndex - 1]
                if !prevTextField.text.isEmpty {
                    prevTextField.text.removeLast()
                    textDidChange(prevTextField)
                }
                prevTextField.becomeFirstResponder(selectionPosition: .end)
                return true
            }
            guard
                !textField.text.isEmpty,
                let selectedRange = textField.selectedRange,
                selectedRange.length < 2
                else { break }
            let symbolIndexToBeDeleted = selectedRange.lowerBound - (selectedRange.length > 0 ? 0 : 1)
            if [cardTextField, dateTextField].contains(textField),
                !textField.text[symbolIndexToBeDeleted].isNumber
            {
                textField.text.remove(at: symbolIndexToBeDeleted - 1)
                textDidChange(textField)
                return true
            }
            
        case #selector(NSResponder.moveLeft):
            guard textField.selectionPosition == .start else { break }
            textFields[safeIndex(currentIndex - 1)].becomeFirstResponder(selectionPosition: .end)
            return true
        
        case #selector(NSResponder.moveRight):
            guard textField.text.isEmpty || textField.selectionPosition == .end else { break }
            textFields[safeIndex(currentIndex + 1)].becomeFirstResponder(selectionPosition: .start)
            return true
            
        case #selector(NSResponder.moveUp):
            guard currentIndex != 0 else { break }
            textFields[0].becomeFirstResponder(selectionPosition: .end)
            return true

        case #selector(NSResponder.moveDown):
            guard currentIndex == 0 else { break }
            textFields[1].becomeFirstResponder(selectionPosition: .end)
            return true
            
        case #selector(NSResponder.cancelOperation):
            cancelButtonTap(textField)
            return true
            
        default: break
        }
        wg_log(.debug, #function, commandSelector)
        return false
    }
}
