// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation

struct BINRange {
    enum CardBrand {
        case visa
        case amex
        case masterCard
        case discover
        case jcb
        case dinersClub
        case unionPay
        
        var maxCVCLength: Int {
            switch self {
            case .amex:
                return 4
                
            default: return 3
            }
        }
    }
    
    private let qRangeLow: String
    private let qRangeHigh: String
    let length: Int
    let brand: CardBrand
    
    static let allRanges: [BINRange] = [
        // American Express
        .init(qRangeLow: "34", qRangeHigh:"34", length: 15, brand: .amex),
        .init(qRangeLow: "37", qRangeHigh:"37", length: 15, brand: .amex),
        
        // Diners Club
        .init(qRangeLow: "30", qRangeHigh:"30", length: 16, brand: .dinersClub),
        .init(qRangeLow: "36", qRangeHigh:"36", length: 14, brand: .dinersClub),
        .init(qRangeLow: "38", qRangeHigh:"39", length: 16, brand: .dinersClub),
        
        // Discover
        .init(qRangeLow: "60", qRangeHigh:"60", length: 16, brand: .discover),
        .init(qRangeLow: "64", qRangeHigh:"65", length: 16, brand: .discover),
        
        // JCB
        .init(qRangeLow: "35", qRangeHigh:"35", length: 16, brand: .jcb),
        
        // Mastercard
        .init(qRangeLow: "50", qRangeHigh:"59", length: 16, brand: .masterCard),
        .init(qRangeLow: "22", qRangeHigh:"27", length: 16, brand: .masterCard),
        .init(qRangeLow: "67", qRangeHigh:"67", length: 16, brand: .masterCard), // Maestro
        
        // UnionPay
        .init(qRangeLow: "62", qRangeHigh:"62", length: 16, brand: .unionPay),
        
        // Visa
        .init(qRangeLow: "40", qRangeHigh:"49", length: 16, brand: .visa),
        .init(qRangeLow: "413600", qRangeHigh:"413600", length: 13, brand: .visa),
        .init(qRangeLow: "444509", qRangeHigh:"444509", length: 13, brand: .visa),
        .init(qRangeLow: "444509", qRangeHigh:"444509", length: 13, brand: .visa),
        .init(qRangeLow: "444550", qRangeHigh:"444550", length: 13, brand: .visa),
        .init(qRangeLow: "450603", qRangeHigh:"450603", length: 13, brand: .visa),
        .init(qRangeLow: "450617", qRangeHigh:"450617", length: 13, brand: .visa),
        .init(qRangeLow: "450628", qRangeHigh:"450629", length: 13, brand: .visa),
        .init(qRangeLow: "450636", qRangeHigh:"450636", length: 13, brand: .visa),
        .init(qRangeLow: "450640", qRangeHigh:"450641", length: 13, brand: .visa),
        .init(qRangeLow: "450662", qRangeHigh:"450662", length: 13, brand: .visa),
        .init(qRangeLow: "463100", qRangeHigh:"463100", length: 13, brand: .visa),
        .init(qRangeLow: "476142", qRangeHigh:"476142", length: 13, brand: .visa),
        .init(qRangeLow: "476143", qRangeHigh:"476143", length: 13, brand: .visa),
        .init(qRangeLow: "492901", qRangeHigh:"492902", length: 13, brand: .visa),
        .init(qRangeLow: "492920", qRangeHigh:"492920", length: 13, brand: .visa),
        .init(qRangeLow: "492923", qRangeHigh:"492923", length: 13, brand: .visa),
        .init(qRangeLow: "492928", qRangeHigh:"492930", length: 13, brand: .visa),
        .init(qRangeLow: "492937", qRangeHigh:"492937", length: 13, brand: .visa),
        .init(qRangeLow: "492939", qRangeHigh:"492939", length: 13, brand: .visa),
        .init(qRangeLow: "492960", qRangeHigh:"492960", length: 13, brand: .visa),
    ]
    
    func matchesNumber(number: String) -> Bool {
        Int(number.prefix(qRangeLow.count))! >= Int(qRangeLow.prefix(number.count))! &&
            Int(number.prefix(qRangeHigh.count))! <= Int(qRangeHigh.prefix(number.count))!
    }
    
    func luhnValidation(number: String) -> PaymentValidator.State {
        guard number.count == length else { return .incomplete }
        let sum = number
            .map { Int(String($0))! }
            .reversed()
            .enumerated()
            .reduce(into: 0) { sum, digit in
                var (i, digit) = digit
                if i % 2 == 1 {
                    digit *= 2
                }
                if digit > 9 {
                    digit -= 9
                }
                sum += digit
        }
        if sum % 10 == 0 {
            return .complete
        }
        return .invalid
    }
    
    static func mostSpecificBINRange(for number: String) -> BINRange? {
        allRanges.filter {
            $0.matchesNumber(number: number)
        }.sorted { $0.length > $1.length }.first
    }
}
