// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Foundation

struct PaymentValidator {
    enum State {
        case incomplete
        case invalid
        
        case complete
        case validFlexible
        
        var valid: Bool {
            switch self {
            case .validFlexible, .complete:
                return true
            case .incomplete, .invalid:
                return false
            }
        }
    }
    enum Kind: CaseIterable {
        case cardNumber
        case date
        case cvc
        case zipCode
    }
    struct Content {
        var card: ConnectionManager.Card
        var cardBrand: BINRange.CardBrand?
        private var state = Kind.allCases.reduce(into: [Kind: State]()) { $0[$1] = .incomplete }
        
        var valid: Bool { state.first { !$0.value.valid } == nil }
        
        mutating func updateState(_ state: State, for kind: Kind) {
            self.state[kind] = state
        }
        
        init(card: ConnectionManager.Card, cardBrand: BINRange.CardBrand?) {
            self.card = card
            self.cardBrand = cardBrand
        }
    }
    
    private(set) var state: State
    private(set) var failed: Bool
    private(set) var outputString: String
    
    static func validate(string: String, kind: Kind, content: inout Content) -> Self {
        let result: Self
        switch kind {
        case .cardNumber:
            result = validateCardNumber(string, content: &content)
            
        case .date:
            result = validateDateString(string, content: &content)
            
        case .cvc:
            result = validateCVCString(string, content: &content)
            
        case .zipCode:
            result = validateZIPString(string, content: &content)
        }
        content.updateState(result.state, for: kind)
        return result
    }
    
    private static func validateCardNumber(_ string: String, content: inout Content) -> Self {
        let sanitizedNumber = string
            .components(separatedBy: CharacterSet.whitespacesAndNewlines)
            .joined()
        var numericString = sanitizedNumber
            .components(separatedBy: CharacterSet.decimalDigits.inverted)
            .joined()
        var result = Self(
            state: .invalid,
            failed: numericString != sanitizedNumber,
            outputString: ""
        )
        if numericString.isEmpty {
            result.state = .incomplete
        } else {
            let binRange = BINRange.mostSpecificBINRange(for: numericString)
            let maxLength = binRange?.length ?? 16
            if numericString.count > maxLength {
                numericString = .init(numericString.prefix(maxLength))
                result.failed = true
            }
            content.cardBrand = binRange?.brand
            result.state = binRange?.luhnValidation(number: numericString) ?? .invalid
        }
        var cardString = numericString
        for i in 0..<(cardString.count - 1) / 4 {
            cardString.insert(" ", at: 5 * i + 4)
        }
        result.outputString = cardString
        content.card.number = numericString
        return result
    }
    
    private static func validateDateString(_ string: String, content: inout Content) -> Self {
        var sanitizedNumber = string
        var shouldInsertZero = false
        if let index = sanitizedNumber.firstIndex(of: "/") {
            sanitizedNumber.remove(at: index)
            shouldInsertZero = index == sanitizedNumber.index(at: 1) && sanitizedNumber[0] != "0"
        }
        let numericString = sanitizedNumber
            .components(separatedBy: CharacterSet.decimalDigits.inverted)
            .joined()
        var result = Self(
            state: .incomplete,
            failed: numericString != sanitizedNumber,
            outputString: numericString
        )
        var month = 0
        var year = 0
        defer {
            content.card.expiryMonth = month
            content.card.expiryYear = year
        }
        guard !numericString.isEmpty else {
            return result
        }
        shouldInsertZero ||= Int(String(result.outputString[0]))! > 3
        if shouldInsertZero {
            result.outputString.insert("0", at: 0)
        }
        guard result.outputString.count > 1 else {
            return result
        }
        let monthString = String(result.outputString[0...1])
        month = Int(monthString)!
        guard month <= 12 else {
            result.outputString = monthString
            result.failed ||= monthString != result.outputString
            result.state = .invalid
            return result
        }
        result.outputString.insert("/", at: 2)
        let expectedSymbolsCount = 5
        if result.outputString.count > expectedSymbolsCount {
            result.failed = true
            result.outputString = .init(result.outputString.prefix(expectedSymbolsCount))
        }
        guard result.outputString.count == expectedSymbolsCount else {
            return result
        }
        let yearString = result.outputString[3...4]
        year = Int(yearString)!
        let currentDateComponents = Calendar.current.dateComponents(.init([.year, .month]), from: .init())
        let shortYear = currentDateComponents.year! % 100
        if year > shortYear
            || year == shortYear && month >= currentDateComponents.month!
        {
            result.state = .complete
        } else {
            result.state = .invalid
        }
        return result
    }
    
    private static func validateCVCString(_ string: String, content: inout Content) -> Self {
        let numericString = string
            .components(separatedBy: CharacterSet.decimalDigits.inverted)
            .joined()
        let validRange = (3...(content.cardBrand?.maxCVCLength ?? 3))
        let state: State
        if string.count < validRange.lowerBound {
            state = .incomplete
        } else if validRange.contains(string.count) {
            state = string.count == validRange.upperBound ? .complete : .validFlexible
        } else {
            state = .invalid
        }
        content.card.cvc = Int(numericString) ?? 0
        return Self(
            state: state,
            failed: numericString != string,
            outputString: numericString
        )
    }
    
    private static func validateZIPString(_ string: String, content: inout Content) -> Self {
        let numericString = string
            .components(separatedBy: CharacterSet.decimalDigits.inverted)
            .joined()
        content.card.addressZip = Int(numericString) ?? 0
        return Self(
            state: numericString.isEmpty ? .incomplete : .validFlexible,
            failed: numericString != string,
            outputString: numericString
        )
    }
}
