// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Cocoa
import Connectivity

class ConnectionWindowController: NSWindowController {
    @IBOutlet weak var networkStateView: NetworkStateView!
    @IBOutlet weak var connectionView: ConnectionView!
    
    private var observers = [AnyObject]()
    
    var connectionManager: ConnectionManager!
    var onLogOut: (() -> Void)?
    
    private var addCardWindowController: AddCardWindowController!
    
    override func windowDidLoad() {
        super.windowDidLoad()
        
        window?.setStaticSize(.appSize)
        window?.backgroundColor = R.color.loginBackground()
        windowFrameAutosaveName = .shared
        
        observers = [
            connectionManager.addStatusChangedHandler(statusHandler: .init({ [weak self] status in
                self?.statusChanged(status)
            })),
            connectionManager.observe(\.error) { [weak self] manager, _ in
                self?.handleError(manager.error)
            },
            NotificationCenter.default.addObserver(
                forName: NSApplication.didBecomeActiveNotification,
                object: nil,
                queue: .main)
            { [weak self] _ in self?.applicationDidBecomeActive() },
        ]
        connectionManager.requestFlexrouterPassword = { [weak self] in
            self?.requestFlexrouterPassword($0)
        }
        connectionManager.initialize()
        connectivity.whenConnected = { [weak self] _ in
            self?.networkStateView.state = .available
        }
        connectivity.framework = .network
        connectivity.startNotifier()
    }
    
    @IBAction func actionButtonTap(_ sender: Any) {
        connectionManager.toggleStatus()
    }
    
    @IBAction func launchBubbleButtonTap(_ sender: Any) {
        connectionManager.createAccountPlanAndLaunch()
    }
    
    @IBAction func myAccountButtonTap(_ sender: Any) {
        NSWorkspace.shared.open(connectionManager.accountURL!)
    }
    
    @IBAction func supportButtonTap(_ sender: Any) {
        NSWorkspace.shared.open(URL(string: "https://support.getbubblenow.com/")!)
    }
    
    @IBAction func addCardButtonTap(_ sender: Any) {
        #if DEVELOPER_ID
        addCardWindowController = .controllerFromNib()
        addCardWindowController.connectionManager = connectionManager
        window?.beginSheet(addCardWindowController.window!) { [weak self] response in
            switch response {
            case .OK:
                self?.connectionManager.createAccountPlanAndLaunch()
                
            case .cancel:
                break
                
            default: fatalError("unhandled sheet response: \(response)")
            }
        }
        #else
        NSWorkspace.shared.open(connectionManager.paymentPageURL!)
        #endif
    }
    
    @IBAction func logOutButtonTap(_ sender: Any) {
        logOut()
    }
    
    func logOut() {
        connectionManager.logOut()
        onLogOut?()
    }
    
    private let connectivity = Connectivity(shouldUseHTTPS: false)
    
    private func statusChanged(_ status: ConnectionManager.Status) {
        connectionView.props = .init(status: status)
    }
    
    private func handleError(_ error: Error?) {
        guard let error = error else { return }
        switch error as? NetworkManagerError {
        case .tokenExpired:
            logOut()
            return
            
        case .networkError(let code) where code == NSURLErrorCancelled:
            if !connectivity.isConnected {
                networkStateView.state = .unavailable
                return
            }
            
        default: break
        }
        (NSApp.delegate as? AppDelegate)?.presentErrorController(error)
    }
    
    private func applicationDidBecomeActive() {
        guard connectionManager.status == .noPaymentMethod else { return }
        connectionManager.checkPaymentMethod()
    }
    
    private var flexrouterPasswordWindowController: FlexrouterPasswordWindowController!
    
    private func requestFlexrouterPassword(_ error: FlexrouterManager.Error?) {
        switch error {
        case .invalidPassword, .none: break
        default:
            handleError(error)
            return
        }
        guard flexrouterPasswordWindowController == nil else {
            flexrouterPasswordWindowController
                .state = error == nil ? .normal : .incorrectPassword
            return
        }
        flexrouterPasswordWindowController = initializeFlexrouterPasswordWindowController()
        window?.beginSheet(flexrouterPasswordWindowController.window!)
        { [weak self] _ in
            self?.flexrouterPasswordWindowController = nil
        }
    }
    
    private func initializeFlexrouterPasswordWindowController()
    -> FlexrouterPasswordWindowController
    {
        let windowController = FlexrouterPasswordWindowController.controllerFromNib()
        windowController.registerHandler = { [weak self, weak windowController] in
            windowController?.state = .registering
            self?.connectionManager.updateFlexrouterPassword($0) {
                guard let windowController = windowController else { return }
                switch $0 {
                case .failure(let error):
                    if case .invalidPassword = error {
                        windowController.state = .incorrectPassword
                    } else {
                        self?.handleError(error)
                        fallthrough
                    }
                    
                case .success:
                    self?.window!.endSheet(windowController.window!, returnCode: .OK)
                }
            }
        }
        return windowController
    }
}

extension ConnectionManager.Status {
    var statusIcon: NSImage? {
        switch self {
        case .active, .restarting:
            return NSImage(named: NSImage.statusAvailableName)
        case .activating, .deactivating:
            return NSImage(named: NSImage.statusPartiallyAvailableName)
        default:
            return NSImage(named: NSImage.statusNoneName)
        }
    }
}

extension ConnectionWindowController: NSWindowDelegate {
    func windowShouldClose(_ sender: NSWindow) -> Bool {
        NSApp.hide(nil)
        return false
    }
}
