// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Cocoa

class ErrorWindowController: NSWindowController {
    @IBOutlet weak var button: GradientButton!
    @IBOutlet weak var titleLabel: NSTextField!
    @IBOutlet weak var shareDetailsView: NSView!
    
    var error: Error?
    private var logsURL: URL?
    private var currentError: Error {
        error ?? NSError(description: "ErrorViewController empty error")
    }
    
    override func windowDidLoad() {
        super.windowDidLoad()
        
        window?.setStaticSize(.appSize)
        window?.backgroundColor = .white
        windowFrameAutosaveName = .shared
        
        button.gradientProps = .contactSupportButton
        wg_log(error: currentError, message: "\(Self.self) \(#function)")
        
        button.isHidden = true
        shareDetailsView.isHidden = true
        Logger.global?.createLogsFile { [weak self] logsURL in
            guard let self = self else { return }
            self.logsURL = logsURL
            let canComposeEmail = self.mailSharingService(logsURL: logsURL) != nil
            self.button.title = canComposeEmail ? "CONTACT SUPPORT" : "SHARE LOGS FILE"
            self.shareDetailsView.isHidden = canComposeEmail
            self.button.isHidden = false
        }
    }
    
    @IBAction func buttonTap(_ sender: Any) {
        do {
            try composeMail(logsURL: logsURL)
        } catch {
            logsURL.map { share(logsURL: $0) }
        }
    }
    
    @IBAction func emailTap(_ sender: Any) {
        enum Buttons: AlertButtons {
            case openMail
            case saveToClipboard
            case cancel
            
            var title: String {
                switch self {
                case .openMail: return "Open Mail app"
                case .saveToClipboard: return "Save to clipboard"
                case .cancel: return "Cancel"
                }
            }
        }
        
        let alert = NSAlert()
        alert.messageText = Constants.supportEmail
        switch alert.runModal(buttonsType: Buttons.self) {
        case .saveToClipboard:
            NSPasteboard.general.declareTypes([.string], owner: nil)
            NSPasteboard.general.setString(Constants.supportEmail, forType: .string)
        case .openMail:
            NSWorkspace.shared.open(URL(string: "mailto:\(Constants.supportEmail)")!)
        case .cancel:
            break
        }
    }
    
    private func composeMail(logsURL: URL?) throws {
        guard
            let (service, items) = mailSharingService(logsURL: logsURL)
            else { throw NSError(description: "Mail app not available") }
        service.subject = "macOS app Bubble error"
        service.recipients = [Constants.supportEmail]
        service.perform(withItems: items)
    }
    
    private func mailSharingService(logsURL: URL?) -> (service: NSSharingService, items: [Any])? {
        let service = NSSharingService(named: .composeEmail)
        let items: [Any] = ["Please describe what happened or just send us the attached logs", logsURL ?? currentError]
        guard service?.canPerform(withItems: items) == true else { return nil }
        return (service!, items)
    }
    
    private func share(logsURL: URL) {
        let sharingPicker = NSSharingServicePicker(items: [logsURL])
        sharingPicker.show(relativeTo: button.frame, of: button.superview!, preferredEdge: .minX)
    }
}
