// SPDX-License-Identifier: AGPL-3.0-only
// Copyright © 2020 Bubble, Inc. All Rights Reserved.

import Cocoa
import SwiftyUserDefaults

class LoginWindowController: NSWindowController {
    enum State {
        case idle
        case waitingForResponse
    }
    
    @IBOutlet weak var userNameTextField: LoginTextField!
    @IBOutlet weak var passwordTextField: LoginTextField!
    @IBOutlet weak var signInButton: NSButton!
    @IBOutlet weak var signUpButton: NSButton!
    @IBOutlet weak var customHostnameButton: NSButton!
    @IBOutlet weak var errorLabel: NSTextField!
    @IBOutlet weak var signUpContainer: NSView!

    private lazy var controls = [
        userNameTextField,
        passwordTextField,
        signInButton,
        signUpButton,
    ]
    
    var onSignIn: ((SageManager) -> Void)?
    var onPrefsRequest: (() -> Void)?
    
    private var defaultsDisposable: DefaultsDisposable?
    private var appPrivacyWindowController: AppPrivacyWindowController?
    
    private var customURL: URL? {
        didSet {
            customHostnameButton.isHidden = customURL == nil
            guard customURL != oldValue else { return }
            error = nil
            guard let customURL = customURL else { return }
            customHostnameButton.title = "Custom hostname: \(customURL)"
        }
    }
    
    override func windowDidLoad() {
        super.windowDidLoad()
        credentials = .loadSaved()
        let firstResponder = [userNameTextField,
                              passwordTextField,
            ].first { $0.stringValue.isEmpty }
        window?.makeFirstResponder(firstResponder)
        userNameTextField.delegate = self
        passwordTextField.delegate = self
        controlTextDidChange(.init(name: .init("")))
        windowFrameAutosaveName = .shared
        window?.setStaticSize(.appSize)
        defaultsDisposable = Defaults
            .observe(\.settingsBundleHostname) { [weak self] change in
                self?.updateCustomURL(change.newValue ?? "")
            }
        updateCustomURL()
        #if !DEVELOPER_ID
        signUpContainer.removeFromSuperview()
        DispatchQueue.main.async { [self] in // hack to show view modally
            showAppPrivacyIfNeeded()
        }
        #endif
    }
    
    @IBAction func signInButtonTap(_ sender: Any) {
        guard let credentials = credentials else { return }
        state = .waitingForResponse
        sageFabric.login(credentials: credentials, customURL: customURL) { [weak self] result in
            DispatchQueue.main.async {
                guard let self = self else { return }
                self.state = .idle
                switch result {
                case .success(let sageManager):
                    self.onSignIn?(sageManager)
                    
                case .failure(let error):
                    self.error = error
                }
            }
        }
    }
    
    @IBAction func signUpButtonTap(_ sender: Any) {
        state = .waitingForResponse
        sageFabric.getRegisterURL { result in
            DispatchQueue.main.async {
                self.state = .idle
                switch result {
                case .success(let url):
                    NSWorkspace.shared.open(url)
                    
                case .failure(let error):
                    self.error = error
                }
            }
        }
    }
    
    @IBAction func customBubbleButtonTap(_ sender: Any) {
        onPrefsRequest?()
    }
    
    private let sageFabric = SageFabric()
    private var credentials: Credentials? {
        get {
            Credentials(username: userNameTextField.stringValue,
                        password: passwordTextField.stringValue)
        }
        set {
            userNameTextField.stringValue = newValue?.username ?? ""
            passwordTextField.stringValue = newValue?.password ?? ""
        }
    }
    private var error: Error? {
        didSet {
            errorLabel.stringValue = error?.fixedLocalizedDescription ?? ""
            
            [userNameTextField,
             passwordTextField,
                ].forEach {
                    $0?.isError = error != nil
            }
        }
    }
    private var state: State = .idle {
        didSet {
            stateUpdated(state)
        }
    }
    
    private func stateUpdated(_ state: State) {
        let controlsEnabled: Bool
        switch state {
        case .idle:
            controlsEnabled = true
            
        case .waitingForResponse:
            controlsEnabled = false
        }
        controls.forEach { $0?.isEnabled = controlsEnabled }
        window?.makeFirstResponder(nil)
        error = nil
    }
    
    private func updateCustomURL(_ value: String? = nil) {
        let hostname = value ?? Defaults.settingsBundleHostname
        customURL = URL(sslSafe: hostname)
    }
    
    private func showAppPrivacyIfNeeded() {
        guard !Defaults.termsAccepted else { return }
        appPrivacyWindowController = AppPrivacyWindowController.controllerFromNib()
        window?.beginSheet(appPrivacyWindowController!.window!) { response in
            Defaults.termsAccepted = response == .OK
        }
    }
}

extension LoginWindowController: NSWindowDelegate {
    func windowShouldClose(_ sender: NSWindow) -> Bool {
        NSApp.terminate(nil)
        return false
    }
}

extension LoginWindowController: NSTextFieldDelegate {
    func controlTextDidBeginEditing(_ obj: Notification) {
        error = nil
    }
    
    func controlTextDidChange(_ obj: Notification) {
        signInButton.isEnabled = credentials != nil
    }
    
    func control(_ control: NSControl, textView: NSTextView, doCommandBy commandSelector: Selector) -> Bool {
        switch commandSelector {
        case #selector(NSResponder.insertNewline(_:)):
            guard credentials != nil else { fallthrough }
            signInButton.performClick(control)
            return true
            
        case #selector(NSResponder.insertTab(_:)):
            guard
                let currentIndex = controls.firstIndex(of: control),
                case let nextIndex = currentIndex + 1,
                nextIndex < controls.count
                else { break }
            switch controls[nextIndex] {
            case let nextControl as TextField:
                nextControl.becomeFirstResponder()
                return true
                
            case let nextControl as Button:
                if credentials != nil {
                    nextControl.performClick(control)
                } else {
                    controls.first??.becomeFirstResponder()
                }
                return true
                
            default: break
            }
            
        default: break
        }
        wg_log(.debug, #function, commandSelector)
        return false
    }
}
